from __future__ import annotations

import functools
import hashlib
import os
import socket
import struct
import sys
import threading
import time
import webbrowser
from collections.abc import Callable
from inspect import Parameter, signature
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .context import context
from .logging import log

if TYPE_CHECKING:
    from .element import Element

_shown_warnings: set[str] = set()

if sys.version_info < (3, 13):
    from asyncio import iscoroutinefunction
else:
    from inspect import iscoroutinefunction


def warn_once(message: str, *, stack_info: bool = False) -> None:
    """Print a warning message only once."""
    if message not in _shown_warnings:
        log.warning(message, stack_info=stack_info)
        _shown_warnings.add(message)


def is_pytest() -> bool:
    """Check if the code is running in pytest."""
    return 'PYTEST_CURRENT_TEST' in os.environ


def is_user_simulation() -> bool:
    """Check if the code is running in with user simulation (see https://nicegui.io/documentation/user)."""
    return 'NICEGUI_USER_SIMULATION' in os.environ


def is_coroutine_function(obj: Any) -> bool:
    """Check if the object is a coroutine function.

    This function is needed because functools.partial is not a coroutine function, but its func attribute is.
    Note: It will return false for coroutine objects.
    """
    while isinstance(obj, functools.partial):
        obj = obj.func
    return iscoroutinefunction(obj)


def expects_arguments(func: Callable) -> bool:
    """Check if the function expects non-variable arguments without a default value."""
    return any(p.default is Parameter.empty and
               p.kind is not Parameter.VAR_POSITIONAL and
               p.kind is not Parameter.VAR_KEYWORD
               for p in signature(func).parameters.values())


def is_file(path: str | Path | None) -> bool:
    """Check if the path is a file that exists."""
    if not path:
        return False
    if isinstance(path, str) and path.strip().startswith('data:'):
        return False  # NOTE: avoid passing data URLs to Path
    try:
        return Path(path).is_file()
    except OSError:
        return False


def hash_file_path(path: Path, *, max_time: float | None = None) -> str:
    """Hash the given path based on its string representation and optionally the last modification time of given files."""
    hasher = hashlib.sha256(path.as_posix().encode())
    if max_time is not None:
        hasher.update(struct.pack('!d', max_time))
    return hasher.hexdigest()[:32]


def is_port_open(host: str, port: int) -> bool:
    """Check if the port is open by checking if a TCP connection can be established."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((host, port))
    except (ConnectionRefusedError, TimeoutError):
        return False
    except Exception:
        return False
    else:
        return True
    finally:
        sock.close()


def schedule_browser(protocol: str, host: str, port: int, path: str) -> tuple[threading.Thread, threading.Event]:
    """Wait non-blockingly for the port to be open, then start a webbrowser.

    This function launches a thread in order to be non-blocking.
    This thread then uses `is_port_open` to check when the port opens.
    When connectivity is confirmed, the webbrowser is launched using `webbrowser.open`.

    The thread is created as a daemon thread, in order to not interfere with Ctrl+C.

    If you need to stop this thread, you can do this by setting the Event, that gets returned.
    The thread will stop with the next loop without opening the browser.

    :return: A tuple consisting of the actual thread object and an event for stopping the thread.
    """
    cancel = threading.Event()

    def in_thread(protocol: str, host: str, port: int, path: str) -> None:
        while not is_port_open(host, port):
            if cancel.is_set():
                return
            time.sleep(0.1)
        webbrowser.open(f'{protocol}://{host}:{port}/{path.lstrip("/")}')

    host = host if host != '0.0.0.0' else '127.0.0.1'
    thread = threading.Thread(target=in_thread, args=(protocol, host, port, path), daemon=True)
    thread.start()
    return thread, cancel


def kebab_to_camel_case(string: str) -> str:
    """Convert a kebab-case string to camelCase."""
    return ''.join(word.capitalize() if i else word for i, word in enumerate(string.split('-')))


def event_type_to_camel_case(string: str) -> str:
    """Convert an event type string to camelCase."""
    return '.'.join(kebab_to_camel_case(part) if part != '-' else part for part in string.split('.'))


def require_top_level_layout(element: Element) -> None:
    """Check if the element is a top level layout element."""
    parent = context.slot.parent
    if parent != parent.client.content:
        raise RuntimeError(
            f'Found top level layout element "{element.__class__.__name__}" inside element "{parent.__class__.__name__}". '
            'Top level layout elements can not be nested but must be direct children of the page content.',
        )



from typing import Any

from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import GenericEventArguments, Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.label_element import LabelElement
from .mixins.validation_element import ValidationDict, ValidationElement, ValidationFunction


class Number(LabelElement, ValidationElement, DisableableElement):
    LOOPBACK = False

    @resolve_defaults
    def __init__(self,
                 label: str | None = DEFAULT_PROP | None, *,
                 placeholder: str | None = DEFAULT_PROP | None,
                 value: float | None = DEFAULT_PROPS['model-value'] | None,
                 min: float | None = DEFAULT_PROP | None,  # pylint: disable=redefined-builtin
                 max: float | None = DEFAULT_PROP | None,  # pylint: disable=redefined-builtin
                 precision: int | None = None,
                 step: float | None = DEFAULT_PROP | None,
                 prefix: str | None = DEFAULT_PROP | None,
                 suffix: str | None = DEFAULT_PROP | None,
                 format: str | None = None,  # pylint: disable=redefined-builtin
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 validation: ValidationFunction | ValidationDict | None = None,
                 ) -> None:
        """Number Input

        This element is based on Quasar's `QInput <https://quasar.dev/vue-components/input>`_ component.

        You can use the `validation` parameter to define a dictionary of validation rules,
        e.g. ``{'Too small!': lambda value: value > 3}``.
        The key of the first rule that fails will be displayed as an error message.
        Alternatively, you can pass a callable that returns an optional error message.
        To disable the automatic validation on every value change, you can use the `without_auto_validation` method.

        :param label: displayed name for the number input
        :param placeholder: text to show if no value is entered
        :param value: the initial value of the field
        :param min: the minimum value allowed
        :param max: the maximum value allowed
        :param precision: the number of decimal places allowed (default: no limit, negative: decimal places before the dot)
        :param step: the step size for the stepper buttons
        :param prefix: a prefix to prepend to the displayed value
        :param suffix: a suffix to append to the displayed value
        :param format: a string like "%.2f" to format the displayed value
        :param on_change: callback to execute when the value changes
        :param validation: dictionary of validation rules or a callable that returns an optional error message (default: None for no validation)
        """
        self.format = format
        super().__init__(tag='q-input', label=label, value=value, on_value_change=on_change, validation=validation)
        self._props['for'] = self.html_id
        self._props['type'] = 'number'
        self._props.set_optional('placeholder', placeholder)
        self._props.set_optional('min', min)
        self._props.set_optional('max', max)
        self._precision = precision
        self._props.set_optional('step', step)
        self._props.set_optional('prefix', prefix)
        self._props.set_optional('suffix', suffix)
        self.on('blur', self.sanitize, [])

    @property
    def min(self) -> float:
        """The minimum value allowed."""
        return self._props.get('min', -float('inf'))

    @min.setter
    def min(self, value: float) -> None:
        if self._props.get('min') == value:
            return
        self._props.set_optional('min', value)
        self.sanitize()

    @property
    def max(self) -> float:
        """The maximum value allowed."""
        return self._props.get('max', float('inf'))

    @max.setter
    def max(self, value: float) -> None:
        if self._props.get('max') == value:
            return
        self._props.set_optional('max', value)
        self.sanitize()

    @property
    def precision(self) -> int | None:
        """The number of decimal places allowed (default: no limit, negative: decimal places before the dot)."""
        return self._precision

    @precision.setter
    def precision(self, value: int | None) -> None:
        self._precision = value
        self.sanitize()

    @property
    def prefix(self) -> str | None:
        """The prefix to prepend to the displayed value.

        *Added in version 3.5.0*
        """
        return self._props.get('prefix')

    @prefix.setter
    def prefix(self, value: str | None) -> None:
        if value is None:
            self._props.pop('prefix', None)
        else:
            self._props['prefix'] = value

    @property
    def suffix(self) -> str | None:
        """The suffix to append to the displayed value.

        *Added in version 3.5.0*
        """
        return self._props.get('suffix')

    @suffix.setter
    def suffix(self, value: str | None) -> None:
        if value is None:
            self._props.pop('suffix', None)
        else:
            self._props['suffix'] = value

    @property
    def out_of_limits(self) -> bool:
        """Whether the current value is out of the allowed limits."""
        return not self.min <= self.value <= self.max

    def sanitize(self) -> None:
        """Sanitize the current value to be within the allowed limits."""
        if self.value is None:
            return
        value = float(self.value)
        value = max(value, self.min)
        value = min(value, self.max)
        if self.precision is not None:
            value = float(round(value, self.precision))
        self.value = float(self.format % value) if self.format else value
        self.update()

    def _event_args_to_value(self, e: GenericEventArguments) -> Any:
        if not e.args:
            return None
        return float(e.args)

    def _value_to_model_value(self, value: Any) -> Any:
        if value is None:
            return None
        if self.format is None:
            old_value = float(self._props.get(self.VALUE_PROP) or 0)
            if old_value == int(old_value) and value == int(value):
                return str(int(value))  # preserve integer representation
            return str(value)
        if value == '':
            return 0
        return self.format % float(value)



from pathlib import Path

from .elements.column import Column as column
from .elements.html import Html as html
from .elements.label import Label as label

SAD_FACE_SVG = (Path(__file__).parent / 'static' / 'sad_face.svg').read_text(encoding='utf-8')


def error_content(status_code: int, exception: str | Exception = '') -> None:
    """Create an error page.

    :param status_code: HTTP status code
    :param exception: exception that caused the error
    """
    if 400 <= status_code <= 499:
        title = "This page doesn't exist."
    elif 500 <= status_code <= 599:
        title = 'Server error'
    else:
        title = 'Unknown error'

    if isinstance(exception, str):
        message = exception
    else:
        message = exception.__class__.__name__
        if str(exception):
            message += ': ' + str(exception)

    with column().style('width: 100%; padding: 5rem 0; align-items: center; gap: 0'):
        html(SAD_FACE_SVG, sanitize=False).style('width: 8rem; padding: 1.25rem 0')
        label(str(status_code)).style('font-size: 3.75rem; line-height: 1; padding: 1.25rem 0')
        label(title).style('font-size: 1.25rem; line-height: 1.75rem; padding: 1.25rem 0')
        label(message).style('font-size: 1.125rem; line-height: 1.75rem; color: rgb(107 114 128)')



import re
from colorsys import rgb_to_yiq
from typing import Any

from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .button import Button as button
from .color_picker import ColorPicker as color_picker
from .mixins.disableable_element import DisableableElement
from .mixins.label_element import LabelElement
from .mixins.value_element import ValueElement

HEX_COLOR_PATTERN_6 = re.compile(r'^#([0-9a-fA-F]{6})$')
HEX_COLOR_PATTERN_3 = re.compile(r'^#([0-9a-fA-F]{3})$')


class ColorInput(LabelElement, ValueElement, DisableableElement):
    LOOPBACK = False

    @resolve_defaults
    def __init__(self,
                 label: str | None = DEFAULT_PROP | None, *,
                 placeholder: str | None = DEFAULT_PROP | None,
                 value: str = DEFAULT_PROPS['model-value'] | '',
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 preview: bool = False,
                 ) -> None:
        """Color Input

        This element extends Quasar's `QInput <https://quasar.dev/vue-components/input>`_ component with a color picker.

        :param label: displayed label for the color input
        :param placeholder: text to show if no color is selected
        :param value: the current color value
        :param on_change: callback to execute when the value changes
        :param preview: change button background to selected color (default: False)
        """
        super().__init__(tag='q-input', label=label, value=value, on_value_change=on_change)
        self._props['for'] = self.html_id
        self._props.set_optional('placeholder', placeholder)

        with self.add_slot('append'):
            self.picker = color_picker(on_pick=lambda e: self.set_value(e.color))
            self.button = button(on_click=self.open_picker, icon='colorize') \
                .props('flat round', remove='color').classes('cursor-pointer')

        self.preview = preview
        self._update_preview()

    def open_picker(self) -> None:
        """Open the color picker"""
        if self.value:
            self.picker.set_color(self.value)
        self.picker.open()

    def _handle_value_change(self, value: Any) -> None:
        super()._handle_value_change(value)
        self._update_preview()

    def _update_preview(self) -> None:
        if not self.preview:
            return

        color = self.value.strip()
        if HEX_COLOR_PATTERN_6.match(color):
            r = int(color[1:3], 16) / 255
            g = int(color[3:5], 16) / 255
            b = int(color[5:7], 16) / 255
        elif HEX_COLOR_PATTERN_3.match(color):
            r = int(color[1], 16) / 15
            g = int(color[2], 16) / 15
            b = int(color[3], 16) / 15
        else:
            self.button.style('background-color: transparent').props(remove='color')
            return
        luminance = rgb_to_yiq(r, g, b)[0]
        icon_color = 'grey-10' if luminance > 0.5 else 'grey-3'
        self.button.style(f'background-color: {color}').props(f'color="{icon_color}"')



from .file_persistent_dict import FilePersistentDict
from .persistent_dict import PersistentDict
from .read_only_dict import ReadOnlyDict
from .redis_persistent_dict import RedisPersistentDict

__all__ = [
    'FilePersistentDict',
    'PersistentDict',
    'ReadOnlyDict',
    'RedisPersistentDict',
]



import contextlib
import gc
import sys
from copy import copy

from starlette.routing import Route

from .. import app, binding, core, dependencies, event, run, ui
from ..client import Client


def prepare_simulation() -> None:
    """Prepare the simulation by adding the run config and setting the storage secret."""
    core.app.config.add_run_config(
        reload=False,
        title='Test App',
        viewport='',
        favicon=None,
        dark=False,
        language='en-US',
        binding_refresh_interval=0.1,
        reconnect_timeout=3.0,
        message_history_length=1000,
        tailwind=True,
        unocss=None,
        prod_js=True,
        show_welcome_message=False,
    )


@contextlib.contextmanager
def nicegui_reset_globals():
    """Reset the global state of the NiceGUI package."""
    for route in list(app.routes):
        if isinstance(route, Route) and (
            not route.path.startswith('/_nicegui/')
            or route.path.startswith('/_nicegui/auto/static')
            or route.path.startswith('/_nicegui/client/')
        ):
            app.remove_route(route.path)

    app.openapi_schema = None
    app.middleware_stack = None
    app.user_middleware.clear()
    app.urls.clear()
    core.reset()

    element_types: list[type[ui.element]] = [ui.element, *_find_all_subclasses(ui.element)]
    default_classes = {t: copy(t._default_classes) for t in element_types}  # pylint: disable=protected-access
    default_styles = {t: copy(t._default_style) for t in element_types}  # pylint: disable=protected-access
    default_props = {t: copy(t._default_props) for t in element_types}  # pylint: disable=protected-access

    dependencies.importmap_overrides.clear()
    Client.instances.clear()
    Client.page_routes.clear()
    Client.shared_head_html = ''
    Client.shared_body_html = ''
    app.reset()
    binding.reset()

    gc.collect()

    try:
        yield
    finally:
        gc.collect()

        app.reset()
        event.reset()
        run.reset()

        for t in element_types:
            t._default_classes = default_classes[t]  # pylint: disable=protected-access
            t._default_style = default_styles[t]  # pylint: disable=protected-access
            t._default_props = default_props[t]  # pylint: disable=protected-access

        for func in Client.page_routes:
            if not func.__module__.startswith('tests.'):
                parts = func.__module__.split('.')
                for i in range(len(parts)):
                    sys.modules.pop('.'.join(parts[:i+1]), None)  # remove the module and all its parents


def _find_all_subclasses(cls: type) -> list[type]:
    subclasses = []
    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(_find_all_subclasses(subclass))
    return subclasses



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Date(ValueElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 value: (
                     str | dict[str, str] | list[str] | list[str | dict[str, str]] | None
                 ) = DEFAULT_PROPS['model-value'] | None,
                 *,
                 mask: str = DEFAULT_PROP | 'YYYY-MM-DD',
                 on_change: Handler[ValueChangeEventArguments] | None = None) -> None:
        """Date Picker

        This element is based on Quasar's `QDate <https://quasar.dev/vue-components/date>`_ component.
        The date is a string in the format defined by the `mask` parameter.

        You can also use the `range` or `multiple` props to select a range of dates or multiple dates::

            ui.date({'from': '2023-01-01', 'to': '2023-01-05'}).props('range')
            ui.date(['2023-01-01', '2023-01-02', '2023-01-03']).props('multiple')
            ui.date([{'from': '2023-01-01', 'to': '2023-01-05'}, '2023-01-07']).props('multiple range')

        :param value: the initial date
        :param mask: the format of the date string (default: 'YYYY-MM-DD')
        :param on_change: callback to execute when changing the date
        """
        super().__init__(tag='q-date', value=value, on_value_change=on_change)
        self._props['mask'] = mask



from typing_extensions import Self

from ..events import ClickEventArguments, Handler, handle_event
from .mixins.disableable_element import DisableableElement
from .mixins.text_element import TextElement


class Item(DisableableElement):

    def __init__(self, text: str = '', *, on_click: Handler[ClickEventArguments] | None = None) -> None:
        """List Item

        Creates a clickable list item based on Quasar's
        `QItem <https://quasar.dev/vue-components/list-and-list-items#qitem-api>`_ component.
        The item should be placed inside a ``ui.list`` or ``ui.menu`` element.
        If the text parameter is provided, an item section will be created with the given text.
        If you want to customize how the text is displayed, you need to create your own item section and label elements.

        :param text: text to be displayed (default: "")
        :param on_click: callback to be executed when clicking on the item (sets the "clickable" prop to True)
        """
        super().__init__(tag='q-item')

        if on_click:
            self.on_click(on_click)

        if text:
            with self:
                ItemSection(text=text)

    def on_click(self, callback: Handler[ClickEventArguments]) -> Self:
        """Add a callback to be invoked when the List Item is clicked."""
        self._props['clickable'] = True  # idempotent
        self.on('click', lambda _: handle_event(callback, ClickEventArguments(sender=self, client=self.client)))
        return self


class ItemSection(TextElement):

    def __init__(self, text: str = '') -> None:
        """List Item Section

        Creates an item section based on Quasar's
        `QItemSection <https://quasar.dev/vue-components/list-and-list-items#qitemsection-api>`_ component.
        The section should be placed inside a ``ui.item`` element.

        :param text: text to be displayed (default: "")
        """
        super().__init__(tag='q-item-section', text=text)


class ItemLabel(TextElement):

    def __init__(self, text: str = '') -> None:
        """List Item Label

        Creates an item label based on Quasar's `QItemLabel <https://quasar.dev/vue-components/list-and-list-items#qitemlabel-api>`_ component.

        :param text: text to be displayed (default: "")
        """
        super().__init__(tag='q-item-label', text=text)



import re
from html.parser import HTMLParser
from pathlib import Path

VOID_TAGS = 'area base br col embed hr img input link meta param source track wbr'.split()


class VBuild:

    def __init__(self, filepath: Path) -> None:
        parser = VueParser(filepath)
        if parser.html is None:
            raise ValueError(f'{filepath} has no template')

        name = filepath.stem.replace('/', '-').replace('\\', '-').replace(':', '-').replace('.', '-')
        html = re.sub(r'^<([\w-]+)', rf'<\1 data-{name}', parser.html)
        self.html = f'<script type="text/x-template" id="tpl-{name}">\n    {html}\n</script>'
        self.style = '\n'.join(add_css_prefix(style, '') for style in parser.styles) + '\n'
        self.style += '\n'.join(add_css_prefix(style, f'*[data-{name}]') for style in parser.scopedStyles)
        self.script = parser.script or ''


class VueParser(HTMLParser):  # pylint: disable=abstract-method  # pylint assumes there is an abstract ``error`` method

    def __init__(self, filepath: Path):
        HTMLParser.__init__(self)
        self.rootTag: str | None = None
        self.html: str | None = None
        self.script: str | None = None
        self.styles: list[str] = []
        self.scopedStyles: list[str] = []
        self._p1: int | None = None
        self._level = 0
        self._tag = ''
        self.feed(filepath.read_text(encoding='utf-8').strip('\n\r\t '))

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        self._tag = tag

        if tag in VOID_TAGS:
            return  # don't manage if it's a void element

        self._level += 1
        if self._level == 1 and tag == 'template':
            if self._p1 is not None:
                raise ValueError('File contains more than one template')
            self._p1 = self._get_offset() + len(self._start_tag_text)
        if self._level == 2 and self._p1:  # test p1, to be sure to be in a template
            if self.rootTag is not None:
                raise ValueError('File has more than one top level tag')
            self.rootTag = tag

    def handle_endtag(self, tag: str) -> None:
        if tag not in VOID_TAGS:
            if tag == 'template' and self._p1:  # don't watch the level (so it can accept malformed HTML)
                self.html = self.rawdata[self._p1:self._get_offset()].strip()
            self._level -= 1

    def handle_data(self, data: str) -> None:
        if self._level == 1:
            if self._tag == 'script':
                self.script = data
            if self._tag == 'style':
                if 'scoped' in self._start_tag_text.lower():
                    self.scopedStyles.append(data)
                else:
                    self.styles.append(data)

    def _get_offset(self) -> int:
        pos = 0
        for _ in range(self.lineno - 1):
            pos = self.rawdata.find('\n', pos) + 1
        return pos + self.offset

    @property
    def _start_tag_text(self) -> str:
        text = self.get_starttag_text()
        assert text is not None
        return text


def add_css_prefix(css: str, prefix: str) -> str:
    """Add the prefix (CSS selector) to all rules in ``css``."""
    media_queries: list[tuple[str, str]] = []
    while '@media' in css:
        p1 = css.find('@media', 0)
        p2 = css.find('{', p1) + 1
        level = 1
        while level > 0:
            level += 1 if css[p2] == '{' else -1 if css[p2] == '}' else 0
            p2 += 1
        block = css[p1:p2]
        media_def = block[:block.find('{')].strip()
        media_css = block[block.find('{') + 1:block.rfind('}')].strip()
        css = css.replace(block, '')
        media_queries.append((media_def, add_css_prefix(media_css, prefix)))

    lines: list[str] = []
    css = re.sub(re.compile(r'/\*.*?\*/', re.DOTALL), '', css)
    css = re.sub(re.compile(r'[ \t\n]+', re.DOTALL), ' ', css)
    for rule in re.findall(r'[^}]+{[^}]+}', css):
        selectors, declarations = rule.split('{', 1)
        if prefix:
            line = [(prefix + ' ' + i.replace(':scope', '').strip()).strip() for i in selectors.split(',')]
        else:
            line = [i.strip() for i in selectors.split(',')]
        lines.append(', '.join(line) + ' {' + declarations.strip())
    lines.extend(f'{d} {{ {c} }}' for d, c in media_queries)
    return '\n'.join(lines).strip()



from __future__ import annotations

import asyncio
import time
import weakref
from collections import deque
from typing import TYPE_CHECKING, Any

from . import background_tasks, core
from .dependencies import JsComponent

if TYPE_CHECKING:
    from .client import Client
    from .element import Element

ElementId = int

ClientId = str
MessageType = str
Payload = Any
Message = tuple[ClientId, MessageType, Payload]

MessageId = int
MessageTime = float
HistoryEntry = tuple[MessageId, MessageTime, Message]


class Deleted:
    """Class for creating a sentinel value for deleted elements."""


deleted = Deleted()


class Outbox:

    def __init__(self, client: Client) -> None:
        self._client = weakref.ref(client)
        self.updates: weakref.WeakValueDictionary[ElementId, Element | Deleted] = weakref.WeakValueDictionary()
        self.messages: deque[Message] = deque()
        self.message_history: deque[HistoryEntry] = deque()
        self.next_message_id: int = 0

        self._loaded_components: set[str] = set()
        self._should_stop = False
        self._enqueue_event: asyncio.Event | None = None

        if core.app.is_started:
            background_tasks.create(self.loop(), name=f'outbox loop {client.id}')
        else:
            core.app.on_startup(self.loop)

    @property
    def client(self) -> Client:
        """The client this outbox belongs to."""
        client = self._client()
        if client is None:
            raise RuntimeError('The client this outbox belongs to has been deleted.')
        return client

    def _set_enqueue_event(self) -> None:
        """Set the enqueue event while accounting for lazy initialization."""
        if self._enqueue_event:
            self._enqueue_event.set()

    def enqueue_update(self, element: Element) -> None:
        """Enqueue an update for the given element."""
        self.client.check_existence()
        self.updates[element.id] = element
        self._set_enqueue_event()

    def enqueue_delete(self, element: Element) -> None:
        """Enqueue a deletion for the given element."""
        self.client.check_existence()
        self.updates[element.id] = deleted
        self._set_enqueue_event()

    def enqueue_message(self, message_type: MessageType, data: Payload, target_id: ClientId) -> None:
        """Enqueue a message for the given client."""
        self.client.check_existence()
        self.messages.append((target_id, message_type, data))
        self._set_enqueue_event()

    async def loop(self) -> None:
        """Send updates and messages to all clients in an endless loop."""
        self._enqueue_event = asyncio.Event()
        self._enqueue_event.set()

        while not self._should_stop:
            try:
                if not self._enqueue_event.is_set():
                    try:
                        await asyncio.wait_for(self._enqueue_event.wait(), timeout=1.0)
                    except (TimeoutError, asyncio.TimeoutError):
                        continue

                client = self.client
                if not client or not client.has_socket_connection:
                    await asyncio.sleep(0.1)
                    continue

                self._enqueue_event.clear()

                coros = []
                if self.updates:
                    data = {
                        element_id: None if element is deleted else element._to_dict()  # type: ignore  # pylint: disable=protected-access
                        for element_id, element in self.updates.items()
                    }
                    js_components = [
                        component
                        for element in self.updates.values()
                        if not isinstance(element, Deleted)
                        and isinstance((component := element.component), JsComponent)
                        and component.name not in self._loaded_components
                    ]
                    if js_components:
                        coros.append(self._emit((client.id, 'load_js_components', {
                            'components': [{'key': c.key, 'tag': c.tag} for c in js_components],
                        })))
                        self._loaded_components.update(c.name for c in js_components)
                    coros.append(self._emit((client.id, 'update', data)))
                    self.updates.clear()

                if self.messages:
                    for message in self.messages:
                        coros.append(self._emit(message))
                    self.messages.clear()

                for coro in coros:
                    try:
                        await coro
                    except Exception as e:
                        core.app.handle_exception(e)
            except asyncio.CancelledError:
                break
            except Exception as e:
                core.app.handle_exception(e)
                await asyncio.sleep(0.1)

    async def _emit(self, message: Message) -> None:
        client_id, message_type, data = message
        data['_id'] = self.next_message_id

        await core.sio.emit(message_type, data, room=client_id)
        if core.air is not None and core.air.is_air_target(client_id):
            await core.air.emit(message_type, data, room=client_id)

        client = self.client
        if client:
            self.message_history.append((self.next_message_id, time.time(), message))
            max_age = core.sio.eio.ping_interval + core.sio.eio.ping_timeout + client.page.resolve_reconnect_timeout()
            while self.message_history and self.message_history[0][1] < time.time() - max_age:
                self.message_history.popleft()
            while len(self.message_history) > core.app.config.message_history_length:
                self.message_history.popleft()

        self.next_message_id += 1

    def try_rewind(self, target_message_id: MessageId) -> None:
        """Rewind to the given message ID and discard all messages before it."""
        # nothing to do, the next message ID is already the target message ID
        if self.next_message_id == target_message_id:
            return

        # rewind to the target message ID
        while self.message_history:
            self.next_message_id, _, message = self.message_history.pop()
            self.messages.appendleft(message)
            if self.next_message_id == target_message_id:
                self.message_history.clear()
                self._set_enqueue_event()
                return

        # target message ID not found, reload the page
        self.client.run_javascript('window.location.reload()')

    def prune_history(self, next_message_id: MessageId) -> None:
        """Prune the message history up to the given message ID."""
        while self.message_history and self.message_history[0][0] < next_message_id:
            self.message_history.popleft()

    def stop(self) -> None:
        """Stop the outbox loop."""
        self._should_stop = True
        self._set_enqueue_event()  # wake the loop so it checks _should_stop immediately



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Time(ValueElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 value: str | None = DEFAULT_PROPS['model-value'] | None, *,
                 mask: str = DEFAULT_PROP | 'HH:mm',
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Time Picker

        This element is based on Quasar's `QTime <https://quasar.dev/vue-components/time>`_ component.
        The time is a string in the format defined by the `mask` parameter.

        :param value: the initial time
        :param mask: the format of the time string (default: 'HH:mm')
        :param on_change: callback to execute when changing the time
        """
        super().__init__(tag='q-time', value=value, on_value_change=on_change)
        self._props['mask'] = mask



from __future__ import annotations

import time
from collections.abc import Callable
from contextlib import suppress
from pathlib import Path
from typing import cast

from typing_extensions import Self

from .. import helpers, optional_features
from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import GenericEventArguments, Handler, MouseEventArguments, handle_event
from ..logging import log
from .image import pil_to_tempfile
from .mixins.content_element import ContentElement
from .mixins.source_element import SourceElement

with suppress(ImportError):
    from PIL.Image import Image as PIL_Image
    optional_features.register('pillow')


class InteractiveImage(SourceElement, ContentElement, component='interactive_image.js'):
    CONTENT_PROP = 'content'
    PIL_CONVERT_FORMAT = 'PNG'

    @resolve_defaults
    def __init__(self,
                 source: str | Path | PIL_Image = '', *,
                 content: str = '',
                 size: tuple[float, float] | None = DEFAULT_PROP | None,
                 on_mouse: Handler[MouseEventArguments] | None = None,
                 events: list[str] = DEFAULT_PROP | ['click'],
                 cross: bool | str = DEFAULT_PROP | False,
                 sanitize: Callable[[str], str] | bool | None = True,  # DEPRECATED: remove `None` in version 4.0.0
                 ) -> None:
        """Interactive Image

        Create an image with an SVG overlay that handles mouse events and yields image coordinates.
        It is also the best choice for non-flickering image updates.
        If the source URL changes faster than images can be loaded by the browser, some images are simply skipped.
        Thereby repeatedly updating the image source will automatically adapt to the available bandwidth.
        See `OpenCV Webcam <https://github.com/zauberzeug/nicegui/tree/main/examples/opencv_webcam/main.py>`_ for an example.

        The mouse event handler is called with mouse event arguments containing

        - `type` (the name of the JavaScript event),
        - `image_x` and `image_y` (image coordinates in pixels),
        - `button` and `buttons` (mouse button numbers from the JavaScript event), as well as
        - `alt`, `ctrl`, `meta`, and `shift` (modifier keys from the JavaScript event).

        You can also pass a tuple of width and height instead of an image source.
        This will create an empty image with the given size.

        :param source: the source of the image; can be an URL, local file path, a base64 string or just an image size
        :param content: SVG content which should be overlaid; viewport has the same dimensions as the image
        :param size: size of the image (width, height) in pixels; only used if `source` is not set
        :param on_mouse: callback for mouse events (contains image coordinates `image_x` and `image_y` in pixels)
        :param events: list of JavaScript events to subscribe to (default: `['click']`)
        :param cross: whether to show crosshairs or a color string (default: `False`)
        :param sanitize: sanitization mode:
            ``True`` (default) uses client-side sanitization via DOMPurify,
            ``False`` disables sanitization (use only with trusted content),
            or pass a callable to apply server-side sanitization
        """
        if sanitize is None:
            helpers.warn_once('`sanitize=None` is deprecated, defaults to `True` and will be removed in version 4.0.0.')
            sanitize = True

        self._sanitize = sanitize
        super().__init__(source=source, content=content)
        self._props['events'] = events[:]
        self._props['cross'] = cross
        self._props['size'] = size
        self._props['sanitize'] = sanitize is True

        if on_mouse:
            self.on_mouse(on_mouse)

    def set_source(self, source: str | Path | PIL_Image) -> None:
        return super().set_source(source)

    def on_mouse(self, on_mouse: Handler[MouseEventArguments]) -> Self:
        """Add a callback to be invoked when a mouse event occurs."""
        def handle_mouse(e: GenericEventArguments) -> None:
            args = cast(dict, e.args)
            arguments = MouseEventArguments(
                sender=self,
                client=self.client,
                type=args.get('mouse_event_type', ''),
                image_x=args.get('image_x', 0.0),
                image_y=args.get('image_y', 0.0),
                button=args.get('button', 0),
                buttons=args.get('buttons', 0),
                alt=args.get('altKey', False),
                ctrl=args.get('ctrlKey', False),
                meta=args.get('metaKey', False),
                shift=args.get('shiftKey', False),
            )
            handle_event(on_mouse, arguments)
        self.on('mouse', handle_mouse)
        return self

    def _set_props(self, source: str | Path | PIL_Image) -> None:
        if optional_features.has('pillow') and isinstance(source, PIL_Image):
            source = pil_to_tempfile(source, self.PIL_CONVERT_FORMAT)
        super()._set_props(source)

    def force_reload(self) -> None:
        """Force the image to reload from the source."""
        if self._props['src'].startswith('data:'):
            log.warning('ui.interactive_image: force_reload() only works with network sources (not data URIs)')
            return
        self._props['t'] = time.time()

    def add_layer(self, *, content: str = '') -> InteractiveImageLayer:
        """Add a new layer with its own content.

        *Added in version 2.17.0*
        """
        with self:
            layer = InteractiveImageLayer(
                source=self.source,
                content=content,
                size=self._props['size'],
                sanitize=self._sanitize,
            ).classes('nicegui-interactive-image-layer')
            self.on('loaded', lambda e: layer.run_method('updateViewbox', e.args['width'], e.args['height']))
            return layer

    def _handle_content_change(self, content: str) -> None:
        if callable(self._sanitize):
            content = self._sanitize(content)
        return super()._handle_content_change(content)


class InteractiveImageLayer(SourceElement, ContentElement, component='interactive_image.js'):
    CONTENT_PROP = 'content'
    PIL_CONVERT_FORMAT = 'PNG'

    def __init__(self, *,
                 source: str,
                 content: str,
                 size: tuple[float, float] | None,
                 sanitize: Callable[[str], str] | bool | None = True,  # DEPRECATED: remove `None` in version 4.0.0
                 ) -> None:
        """Interactive Image Layer

        This element is created when adding a layer to an ``InteractiveImage``.

        *Added in version 2.17.0*
        """
        if sanitize is None:
            helpers.warn_once('`sanitize=None` is deprecated, defaults to `True` and will be removed in version 4.0.0.')
            sanitize = True
        self._sanitize = sanitize
        super().__init__(source=source, content=content)
        self._props['size'] = size
        self._props['sanitize'] = sanitize is True

    def _set_props(self, source: str | Path | PIL_Image) -> None:
        if optional_features.has('pillow') and isinstance(source, PIL_Image):
            source = pil_to_tempfile(source, self.PIL_CONVERT_FORMAT)
        super()._set_props(source)

    def _handle_content_change(self, content: str) -> None:
        if callable(self._sanitize):
            content = self._sanitize(content)
        return super()._handle_content_change(content)



from typing import Literal

from ..defaults import DEFAULT_PROP, resolve_defaults
from ..element import Element


class Skeleton(Element):

    @resolve_defaults
    def __init__(self,
                 type: Literal[  # pylint: disable=redefined-builtin
                     'text',
                     'rect',
                     'circle',
                     'QBtn',
                     'QBadge',
                     'QChip',
                     'QToolbar',
                     'QCheckbox',
                     'QRadio',
                     'QToggle',
                     'QSlider',
                     'QRange',
                     'QInput',
                     'QAvatar',
                 ] = DEFAULT_PROP | 'rect',
                 *,
                 tag: str = DEFAULT_PROP | 'div',
                 animation: Literal[
                     'wave',
                     'pulse',
                     'pulse-x',
                     'pulse-y',
                     'fade',
                     'blink',
                     'none',
                 ] = DEFAULT_PROP | 'wave',
                 animation_speed: float = 1.5,
                 square: bool = DEFAULT_PROP | False,
                 bordered: bool = DEFAULT_PROP | False,
                 size: str | None = DEFAULT_PROP | None,
                 width: str | None = DEFAULT_PROP | None,
                 height: str | None = DEFAULT_PROP | None,
                 ) -> None:
        """Skeleton

        This element is based on Quasar's `QSkeleton <https://quasar.dev/vue-components/skeleton>`_ component.
        It serves as a placeholder for loading content in cards, menus and other component containers.
        See the `Quasar documentation <https://quasar.dev/vue-components/skeleton/#predefined-types>`_ for a list of available types.

        :param type: type of skeleton to display (default: "rect")
        :param tag: HTML tag to use for this element (default: "div")
        :param animation: animation effect of the skeleton placeholder (default: "wave")
        :param animation_speed: animation speed in seconds (default: 1.5)
        :param square: whether to remover border-radius so borders are squared (default: ``False``)
        :param bordered: whether to apply a default border to the component (default: ``False``)
        :param size: size in CSS units (overrides ``width`` and ``height``)
        :param width: width in CSS units (overridden by ``size`` if set)
        :param height: height in CSS units (overridden by ``size`` if set)
        """
        super().__init__('q-skeleton')
        self._props.set_optional('type', type if type != 'rect' else None)
        self._props.set_optional('tag', tag if tag != 'div' else None)
        self._props.set_optional('animation', animation if animation != 'wave' else None)
        self._props.set_optional('animation-speed', animation_speed if animation_speed != 1.5 else None)
        self._props.set_bool('square', square)
        self._props.set_bool('bordered', bordered)
        self._props.set_optional('size', size)
        self._props.set_optional('width', width)
        self._props.set_optional('height', height)



from collections.abc import Sequence

from ..context import context
from ..events import GenericEventArguments, Handler


def on(type: str,  # pylint: disable=redefined-builtin
       handler: Handler[GenericEventArguments] | None = None,
       args: None | Sequence[str] | Sequence[Sequence[str] | None] = None, *,
       throttle: float = 0.0,
       leading_events: bool = True,
       trailing_events: bool = True,
       ):
    """Subscribe to a global event.

    :param type: name of the event
    :param handler: callback that is called upon occurrence of the event
    :param args: arguments included in the event message sent to the event handler (default: `None` meaning all)
    :param throttle: minimum time (in seconds) between event occurrences (default: 0.0)
    :param leading_events: whether to trigger the event handler immediately upon the first event occurrence (default: `True`)
    :param trailing_events: whether to trigger the event handler after the last event occurrence (default: `True`)
    """
    context.client.layout.on(type, handler, args,
                             throttle=throttle, leading_events=leading_events, trailing_events=trailing_events)



from typing import TYPE_CHECKING, Any, Literal

from typing_extensions import Self

from .. import optional_features
from ..defaults import DEFAULT_PROP, resolve_defaults
from ..element import Element
from ..events import (
    GenericEventArguments,
    Handler,
    TableSelectionEventArguments,
    ValueChangeEventArguments,
    handle_event,
)
from ..logging import log
from .mixins.filter_element import FilterElement

optional_features.try_register('pandas')
if TYPE_CHECKING:
    import pandas as pd

optional_features.try_register('polars')
if TYPE_CHECKING:
    import polars as pl


class Table(FilterElement, component='table.js'):

    @resolve_defaults
    def __init__(self,
                 *,
                 rows: list[dict],
                 columns: list[dict] | None = None,
                 column_defaults: dict | None = None,
                 row_key: str = DEFAULT_PROP | 'id',
                 title: str | None = DEFAULT_PROP | None,
                 selection: Literal[None, 'single', 'multiple'] = DEFAULT_PROP | None,
                 pagination: int | dict | None = None,
                 on_select: Handler[TableSelectionEventArguments] | None = None,
                 on_pagination_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Table

        A table based on Quasar's `QTable <https://quasar.dev/vue-components/table>`_ component.
        Updates can be pushed to the table by updating the ``rows`` or ``columns`` properties.

        If ``selection`` is "single" or "multiple", then a ``selected`` property is accessible containing the selected rows.

        Note:
        Cells in ``rows`` must not contain lists because they can cause the browser to crash.
        To display complex data structures, convert them to strings first (e.g., using ``str()`` or custom formatting).

        :param rows: list of row objects
        :param columns: list of column objects (defaults to the columns of the first row *since version 2.0.0*)
        :param column_defaults: optional default column properties, *added in version 2.0.0*
        :param row_key: name of the column containing unique data identifying the row (default: "id")
        :param title: title of the table
        :param selection: selection type ("single" or "multiple"; default: `None`)
        :param pagination: a dictionary correlating to a pagination object or number of rows per page (`None` hides the pagination, 0 means "infinite"; default: `None`).
        :param on_select: callback which is invoked when the selection changes
        :param on_pagination_change: callback which is invoked when the pagination changes
        """
        super().__init__()

        if columns is None:
            first_row = rows[0] if rows else {}
            columns = [{'name': key, 'label': str(key).upper(), 'field': key, 'sortable': True} for key in first_row]

        self._column_defaults = column_defaults
        self._use_columns_from_df = False
        self._props['columns'] = self._normalize_columns(columns)
        self._props['rows'] = rows
        self._props['row-key'] = row_key
        self._props.set_optional('title', title)
        self._props['hide-pagination'] = pagination is None
        self._props['pagination'] = pagination if isinstance(pagination, dict) else {'rowsPerPage': pagination or 0}
        self._props['selection'] = selection or 'none'
        self._props['selected'] = []
        self._props['fullscreen'] = False
        self._selection_handlers = [on_select] if on_select else []
        self._pagination_change_handlers = [on_pagination_change] if on_pagination_change else []

        def handle_selection(e: GenericEventArguments) -> None:
            if e.args['added']:
                if self.selection == 'single':
                    self.selected.clear()
                self.selected.extend(e.args['rows'])
            else:
                self.selected = [row for row in self.selected if row[self.row_key] not in e.args['keys']]
            arguments = TableSelectionEventArguments(sender=self, client=self.client, selection=self.selected)
            for handler in self._selection_handlers:
                handle_event(handler, arguments)
        self.on('selection', handle_selection, ['added', 'rows', 'keys'])

        def handle_pagination_change(e: GenericEventArguments) -> None:
            previous_value = self.pagination
            self.pagination = e.args
            arguments = ValueChangeEventArguments(sender=self, client=self.client,
                                                  value=self.pagination, previous_value=previous_value)
            for handler in self._pagination_change_handlers:
                handle_event(handler, arguments)
        self.on('update:pagination', handle_pagination_change)

    def _to_dict(self) -> dict[str, Any]:
        # scan rows for lists and add slot templates if needed
        for column in self._props['columns']:
            field = column.get('field')
            name = column.get('name')
            if not field or not name or f'body-cell-{name}' in self.slots:
                continue
            for row in self._props['rows']:
                value = row.get(field)
                if isinstance(value, (list, set, tuple)):
                    log.warning(
                        f'Found list in column "{name}": {value}.\n'
                        'Unless there is slot template, table rows must not contain lists or the browser will crash.\n'
                        'NiceGUI is intervening by adding a slot template to display the list as comma-separated values.'
                    )
                    self.add_slot(f'body-cell-{name}', '''
                        <td class="text-right" :props="props">
                            {{ Array.isArray(props.value) ? props.value.join(', ') : props.value }}
                        </td>
                    ''')
                    break

        return super()._to_dict()

    def on_select(self, callback: Handler[TableSelectionEventArguments]) -> Self:
        """Add a callback to be invoked when the selection changes."""
        self._selection_handlers.append(callback)
        return self

    def on_pagination_change(self, callback: Handler[ValueChangeEventArguments]) -> Self:
        """Add a callback to be invoked when the pagination changes."""
        self._pagination_change_handlers.append(callback)
        return self

    def _normalize_columns(self, columns: list[dict]) -> list[dict]:
        return [{**self._column_defaults, **column} for column in columns] if self._column_defaults else columns

    @classmethod
    def from_pandas(cls,
                    df: 'pd.DataFrame', *,
                    columns: list[dict] | None = None,
                    column_defaults: dict | None = None,
                    row_key: str = 'id',
                    title: str | None = None,
                    selection: Literal['single', 'multiple'] | None = None,
                    pagination: int | dict | None = None,
                    on_select: Handler[TableSelectionEventArguments] | None = None) -> Self:
        """Create a table from a Pandas DataFrame.

        Note:
        If the DataFrame contains non-serializable columns of type `datetime64[ns]`, `timedelta64[ns]`, `complex128` or `period[M]`,
        they will be converted to strings.
        To use a different conversion, convert the DataFrame manually before passing it to this method.
        See `issue 1698 <https://github.com/zauberzeug/nicegui/issues/1698>`_ for more information.

        *Added in version 2.0.0*

        :param df: Pandas DataFrame
        :param columns: list of column objects (defaults to the columns of the dataframe)
        :param column_defaults: optional default column properties
        :param row_key: name of the column containing unique data identifying the row (default: "id")
        :param title: title of the table
        :param selection: selection type ("single" or "multiple"; default: `None`)
        :param pagination: a dictionary correlating to a pagination object or number of rows per page (`None` hides the pagination, 0 means "infinite"; default: `None`).
        :param on_select: callback which is invoked when the selection changes
        :return: table element
        """
        rows, columns_from_df = cls._pandas_df_to_rows_and_columns(df)
        table = cls(
            rows=rows,
            columns=columns or columns_from_df,
            column_defaults=column_defaults,
            row_key=row_key,
            title=title,
            selection=selection,
            pagination=pagination,
            on_select=on_select,
        )
        table._use_columns_from_df = columns is None
        return table

    @classmethod
    def from_polars(cls,
                    df: 'pl.DataFrame', *,
                    columns: list[dict] | None = None,
                    column_defaults: dict | None = None,
                    row_key: str = 'id',
                    title: str | None = None,
                    selection: Literal['single', 'multiple'] | None = None,
                    pagination: int | dict | None = None,
                    on_select: Handler[TableSelectionEventArguments] | None = None) -> Self:
        """Create a table from a Polars DataFrame.

        Note:
        If the DataFrame contains non-UTF-8 datatypes, they will be converted to strings.
        To use a different conversion, convert the DataFrame manually before passing it to this method.

        *Added in version 2.7.0*

        :param df: Polars DataFrame
        :param columns: list of column objects (defaults to the columns of the dataframe)
        :param column_defaults: optional default column properties
        :param row_key: name of the column containing unique data identifying the row (default: "id")
        :param title: title of the table
        :param selection: selection type ("single" or "multiple"; default: `None`)
        :param pagination: a dictionary correlating to a pagination object or number of rows per page (`None` hides the pagination, 0 means "infinite"; default: `None`).
        :param on_select: callback which is invoked when the selection changes
        :return: table element
        """
        rows, columns_from_df = cls._polars_df_to_rows_and_columns(df)
        table = cls(
            rows=rows,
            columns=columns or columns_from_df,
            column_defaults=column_defaults,
            row_key=row_key,
            title=title,
            selection=selection,
            pagination=pagination,
            on_select=on_select,
        )
        table._use_columns_from_df = columns is None
        return table

    def update_from_pandas(self,
                           df: 'pd.DataFrame', *,
                           clear_selection: bool = True,
                           columns: list[dict] | None = None,
                           column_defaults: dict | None = None) -> None:
        """Update the table from a Pandas DataFrame.

        See `from_pandas()` for more information about the conversion of non-serializable columns.

        If `columns` is not provided and the columns had been inferred from a DataFrame,
        the columns will be updated to match the new DataFrame.

        :param df: Pandas DataFrame
        :param clear_selection: whether to clear the selection (default: True)
        :param columns: list of column objects (defaults to the columns of the dataframe)
        :param column_defaults: optional default column properties
        """
        rows, columns_from_df = self._pandas_df_to_rows_and_columns(df)
        self._update_table(rows, columns_from_df, clear_selection, columns, column_defaults)

    def update_from_polars(self,
                           df: 'pl.DataFrame', *,
                           clear_selection: bool = True,
                           columns: list[dict] | None = None,
                           column_defaults: dict | None = None) -> None:
        """Update the table from a Polars DataFrame.

        :param df: Polars DataFrame
        :param clear_selection: whether to clear the selection (default: True)
        :param columns: list of column objects (defaults to the columns of the dataframe)
        :param column_defaults: optional default column properties
        """
        rows, columns_from_df = self._polars_df_to_rows_and_columns(df)
        self._update_table(rows, columns_from_df, clear_selection, columns, column_defaults)

    def _update_table(self,
                      rows: list[dict],
                      columns_from_df: list[dict],
                      clear_selection: bool,
                      columns: list[dict] | None,
                      column_defaults: dict | None) -> None:
        """Helper function to update the table."""
        self.rows[:] = rows
        if column_defaults is not None:
            self._column_defaults = column_defaults
        if columns or self._use_columns_from_df:
            self.columns[:] = self._normalize_columns(columns or columns_from_df)
        if clear_selection:
            self.selected.clear()

    @staticmethod
    def _pandas_df_to_rows_and_columns(df: 'pd.DataFrame') -> tuple[list[dict], list[dict]]:
        import pandas as pd  # pylint: disable=import-outside-toplevel

        def is_special_dtype(dtype):
            return (pd.api.types.is_datetime64_any_dtype(dtype) or
                    pd.api.types.is_timedelta64_dtype(dtype) or
                    pd.api.types.is_complex_dtype(dtype) or
                    pd.api.types.is_object_dtype(dtype) or
                    isinstance(dtype, (pd.PeriodDtype, pd.IntervalDtype)))
        special_cols = df.columns[df.dtypes.apply(is_special_dtype)]
        if not special_cols.empty:
            df = df.copy()
            df[special_cols] = df[special_cols].astype(str)

        if isinstance(df.columns, pd.MultiIndex):
            raise ValueError('MultiIndex columns are not supported. '
                             'You can convert them to strings using something like '
                             '`df.columns = ["_".join(col) for col in df.columns.values]`.')

        return df.to_dict('records'), [{'name': col, 'label': col, 'field': col} for col in df.columns]

    @staticmethod
    def _polars_df_to_rows_and_columns(df: 'pl.DataFrame') -> tuple[list[dict], list[dict]]:
        return df.to_dicts(), [{'name': col, 'label': col, 'field': col} for col in df.columns]

    @property
    def rows(self) -> list[dict]:
        """List of rows."""
        return self._props['rows']

    @rows.setter
    def rows(self, value: list[dict]) -> None:
        self._props['rows'] = value

    @property
    def columns(self) -> list[dict]:
        """List of columns."""
        return self._props['columns']

    @columns.setter
    def columns(self, value: list[dict]) -> None:
        self._props['columns'] = self._normalize_columns(value)

    @property
    def column_defaults(self) -> dict | None:
        """Default column properties."""
        return self._column_defaults

    @column_defaults.setter
    def column_defaults(self, value: dict | None) -> None:
        self._column_defaults = value
        self.columns = self.columns  # re-normalize columns

    @property
    def row_key(self) -> str:
        """Name of the column containing unique data identifying the row."""
        return self._props['row-key']

    @row_key.setter
    def row_key(self, value: str) -> None:
        self._props['row-key'] = value

    @property
    def selected(self) -> list[dict]:
        """List of selected rows."""
        return self._props['selected']

    @selected.setter
    def selected(self, value: list[dict]) -> None:
        self._props['selected'] = value

    @property
    def selection(self) -> Literal[None, 'single', 'multiple']:
        """Selection type.

        *Added in version 2.11.0*
        """
        return None if self._props['selection'] == 'none' else self._props['selection']

    @selection.setter
    def selection(self, value: Literal[None, 'single', 'multiple']) -> None:
        self._props['selection'] = value or 'none'

    def set_selection(self, value: Literal[None, 'single', 'multiple']) -> None:
        """Set the selection type.

        *Added in version 2.11.0*
        """
        self.selection = value

    @property
    def pagination(self) -> dict:
        """Pagination object."""
        return self._props['pagination']

    @pagination.setter
    def pagination(self, value: dict) -> None:
        self._props['pagination'] = value

    @property
    def is_fullscreen(self) -> bool:
        """Whether the table is in fullscreen mode."""
        return self._props['fullscreen']

    @is_fullscreen.setter
    def is_fullscreen(self, value: bool) -> None:
        """Set fullscreen mode."""
        self._props['fullscreen'] = value

    def set_fullscreen(self, value: bool) -> None:
        """Set fullscreen mode."""
        self.is_fullscreen = value

    def toggle_fullscreen(self) -> None:
        """Toggle fullscreen mode."""
        self.is_fullscreen = not self.is_fullscreen

    def add_rows(self, rows: list[dict]) -> None:
        """Add rows to the table."""
        self.rows.extend(rows)

    def add_row(self, row: dict) -> None:
        """Add a single row to the table."""
        self.add_rows([row])

    def remove_rows(self, rows: list[dict]) -> None:
        """Remove rows from the table."""
        keys = [row[self.row_key] for row in rows]
        self.rows[:] = [row for row in self.rows if row[self.row_key] not in keys]
        self.selected[:] = [row for row in self.selected if row[self.row_key] not in keys]

    def remove_row(self, row: dict) -> None:
        """Remove a single row from the table."""
        self.remove_rows([row])

    def update_rows(self, rows: list[dict], *, clear_selection: bool = True) -> None:
        """Update rows in the table.

        :param rows: list of rows to update
        :param clear_selection: whether to clear the selection (default: True)
        """
        self.rows[:] = rows
        if clear_selection:
            self.selected.clear()

    async def get_filtered_sorted_rows(self, *, timeout: float = 1) -> list[dict]:
        """Asynchronously return the filtered and sorted rows of the table."""
        return await self.get_computed_prop('filteredSortedRows', timeout=timeout)

    async def get_computed_rows(self, *, timeout: float = 1) -> list[dict]:
        """Asynchronously return the computed rows of the table."""
        return await self.get_computed_prop('computedRows', timeout=timeout)

    async def get_computed_rows_number(self, *, timeout: float = 1) -> int:
        """Asynchronously return the number of computed rows of the table."""
        return await self.get_computed_prop('computedRowsNumber', timeout=timeout)

    class row(Element):

        def __init__(self) -> None:
            """Row Element

            This element is based on Quasar's `QTr <https://quasar.dev/vue-components/table#qtr-api>`_ component.
            """
            super().__init__('q-tr')

    class header(Element, default_classes='[&>*]:inline-block'):

        def __init__(self, column_name: str | None = None) -> None:
            """Header Element

            This element is based on Quasar's `QTh <https://quasar.dev/vue-components/table#qth-api>`_ component.

            :param column_name: corresponding column to access alignment and other properties (*added in version 3.5.0*)
            """
            super().__init__('q-th')
            if column_name is not None:
                self._props[':props'] = 'props'
                self._props['key'] = column_name

    class cell(Element):

        def __init__(self, column_name: str | None = None) -> None:
            """Cell Element

            This element is based on Quasar's `QTd <https://quasar.dev/vue-components/table#qtd-api>`_ component.

            :param column_name: corresponding column to access alignment and other properties (*added in version 3.5.0*)
            """
            super().__init__('q-td')
            if column_name is not None:
                self._props[':props'] = 'props'
                self._props['key'] = column_name



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Splitter(ValueElement, DisableableElement, default_classes='nicegui-splitter'):

    @resolve_defaults
    def __init__(self, *,
                 horizontal: bool | None = DEFAULT_PROP | False,
                 reverse: bool | None = DEFAULT_PROP | False,
                 limits: tuple[float, float] | None = DEFAULT_PROP | (0, 100),
                 value: float | None = DEFAULT_PROPS['model-value'] | 50,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Splitter

        The `ui.splitter` element divides the screen space into resizable sections,
        allowing for flexible and responsive layouts in your application.

        Based on Quasar's Splitter component:
        `Splitter <https://quasar.dev/vue-components/splitter>`_

        It provides three customizable slots, ``before``, ``after``, and ``separator``,
        which can be used to embed other elements within the splitter.

        :param horizontal: Whether to split horizontally instead of vertically
        :param limits: Two numbers representing the minimum and maximum split size of the two panels
        :param value: Size of the first panel (or second if using reverse)
        :param reverse: Whether to apply the model size to the second panel instead of the first
        :param on_change: callback which is invoked when the user releases the splitter
        """
        super().__init__(tag='q-splitter', value=value, on_value_change=on_change, throttle=0.05)
        self._props.set_bool('horizontal', horizontal)
        self._props.set_bool('reverse', reverse)
        self._props.set_optional('limits', limits)

        self.before = self.add_slot('before')
        self.after = self.add_slot('after')
        self.separator = self.add_slot('separator')



from __future__ import annotations

import asyncio
import copyreg
import dataclasses
import time
import weakref
from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Literal, TypeVar

from typing_extensions import dataclass_transform

from . import core
from .logging import log

if TYPE_CHECKING:
    from _typeshed import DataclassInstance, IdentityFunction

MAX_PROPAGATION_TIME = 0.01

propagation_visited: ContextVar[set[tuple[int, str]] | None] = ContextVar('propagation_visited', default=None)

bindings: defaultdict[tuple[int, str], list[tuple[Any, Any, str, Callable[[Any], Any] | None]]] = defaultdict(list)
bindable_properties: weakref.WeakValueDictionary[tuple[int, str], Any] = weakref.WeakValueDictionary()
active_links: list[tuple[Any, str, Any, str, Callable[[Any], Any] | None]] = []
_active_links_added = asyncio.Event()

TC = TypeVar('TC', bound=type)
T = TypeVar('T')


def _has_attribute(obj: object | Mapping, name: str) -> Any:
    if isinstance(obj, Mapping):
        return name in obj
    return hasattr(obj, name)


def _get_attribute(obj: object | Mapping, name: str) -> Any:
    if isinstance(obj, Mapping):
        return obj[name]
    return getattr(obj, name)


def _set_attribute(obj: object | Mapping, name: str, value: Any) -> None:
    if isinstance(obj, dict):
        obj[name] = value
    else:
        setattr(obj, name, value)


async def refresh_loop() -> None:
    """Refresh all bindings in an endless loop."""
    global _active_links_added  # pylint: disable=global-statement # noqa: PLW0603
    _active_links_added = asyncio.Event()
    await _active_links_added.wait()
    if core.app.config.binding_refresh_interval is None:
        core.app.config.binding_refresh_interval = 0.1
        log.warning('Starting active binding loop even though it was disabled via binding_refresh_interval=None.')
    while True:
        _refresh_step()
        try:
            await asyncio.sleep(core.app.config.binding_refresh_interval)
        except asyncio.CancelledError:
            break


def _refresh_step() -> None:
    t = time.time()
    for link in active_links:
        (source_obj, source_name, target_obj, target_name, transform) = link
        if _has_attribute(source_obj, source_name):
            source_value = _get_attribute(source_obj, source_name)
            value = transform(source_value) if transform else source_value
            if not _has_attribute(target_obj, target_name) or _get_attribute(target_obj, target_name) != value:
                _set_attribute(target_obj, target_name, value)
                _propagate(target_obj, target_name)
        del link, source_obj, target_obj  # pylint: disable=modified-iterating-list
    if time.time() - t > MAX_PROPAGATION_TIME:
        log.warning(f'binding propagation for {len(active_links)} active links took {time.time() - t:.3f} s')


def _propagate(source_obj: Any, source_name: str) -> None:
    token = propagation_visited.set(set())
    try:
        _propagate_recursively(source_obj, source_name)
    finally:
        propagation_visited.reset(token)


def _propagate_recursively(source_obj: Any, source_name: str) -> None:
    visited = propagation_visited.get()
    assert visited is not None, 'propagation_visited is not set'

    source_obj_id = id(source_obj)
    if (source_obj_id, source_name) in visited:
        return
    visited.add((source_obj_id, source_name))

    if not _has_attribute(source_obj, source_name):
        return
    source_value = _get_attribute(source_obj, source_name)

    for _, target_obj, target_name, transform in bindings.get((source_obj_id, source_name), []):
        if (id(target_obj), target_name) in visited:
            continue

        target_value = transform(source_value) if transform else source_value
        if not _has_attribute(target_obj, target_name) or _get_attribute(target_obj, target_name) != target_value:
            _set_attribute(target_obj, target_name, target_value)
            _propagate_recursively(target_obj, target_name)


def _check_attribute_exists(other_obj: Any, other_name: str, *, role: Literal['self', 'other']) -> None:
    if not _has_attribute(other_obj, other_name):
        if isinstance(other_obj, Mapping):
            raise KeyError(
                f'Could not bind non-existing key "{other_name}". '
                f'To allow missing keys (lazy binding), remove {role}_strict=True or add the key before binding.'
            )
        raise AttributeError(
            f'Could not bind non-existing attribute "{other_name}" on object of type {other_obj.__class__.__name__}. '
            f'To allow missing attributes (lazy binding), add {role}_strict=False or add the attribute before binding.'
        )


def _check_self_and_other_attribute(self_obj: Any, self_name: str, other_obj: Any, other_name: str,
                                    self_strict: bool | None, other_strict: bool | None) -> None:
    if self_strict or (self_strict is None and not isinstance(self_obj, dict)):
        _check_attribute_exists(self_obj, self_name, role='self')
    if other_strict or (other_strict is None and not isinstance(other_obj, dict)):
        _check_attribute_exists(other_obj, other_name, role='other')


def bind_to(self_obj: Any, self_name: str, other_obj: Any, other_name: str,
            forward: Callable[[Any], Any] | None = None, *,
            self_strict: bool | None = None, other_strict: bool | None = None) -> None:
    """Bind the property of one object to the property of another object.

    The binding works one way only, from the first object to the second.
    The update happens immediately and whenever a value changes.

    :param self_obj: The object to bind from.
    :param self_name: The name of the property to bind from.
    :param other_obj: The object to bind to.
    :param other_name: The name of the property to bind to.
    :param forward: A function to apply to the value before applying it (default: identity).
    :param self_strict: Whether to check (and raise) if the first object has the specified property
        (default: None, performs a check if the object is not a dictionary, *added in version 3.0.0*).
    :param other_strict: Whether to check (and raise) if the second object has the specified property
        (default: None, performs a check if the object is not a dictionary, *added in version 3.0.0*).
    """
    _check_self_and_other_attribute(self_obj, self_name, other_obj, other_name, self_strict, other_strict)
    bindings[(id(self_obj), self_name)].append((self_obj, other_obj, other_name, forward))
    if (id(self_obj), self_name) not in bindable_properties:
        active_links.append((self_obj, self_name, other_obj, other_name, forward))
        _active_links_added.set()
    _propagate(self_obj, self_name)


def bind_from(self_obj: Any, self_name: str, other_obj: Any, other_name: str,
              backward: Callable[[Any], Any] | None = None, *,
              self_strict: bool | None = None, other_strict: bool | None = None) -> None:
    """Bind the property of one object from the property of another object.

    The binding works one way only, from the second object to the first.
    The update happens immediately and whenever a value changes.

    :param self_obj: The object to bind to.
    :param self_name: The name of the property to bind to.
    :param other_obj: The object to bind from.
    :param other_name: The name of the property to bind from.
    :param backward: A function to apply to the value before applying it (default: identity).
    :param self_strict: Whether to check (and raise) if the first object has the specified property (default: None,
        performs a check if the object is not a dictionary, *added in version 3.0.0*).
    :param other_strict: Whether to check (and raise) if the second object has the specified property (default: None,
        performs a check if the object is not a dictionary, *added in version 3.0.0*).
    """
    _check_self_and_other_attribute(self_obj, self_name, other_obj, other_name, self_strict, other_strict)
    bindings[(id(other_obj), other_name)].append((other_obj, self_obj, self_name, backward))
    if (id(other_obj), other_name) not in bindable_properties:
        active_links.append((other_obj, other_name, self_obj, self_name, backward))
        _active_links_added.set()
    _propagate(other_obj, other_name)


def bind(self_obj: Any, self_name: str, other_obj: Any, other_name: str, *,
         forward: Callable[[Any], Any] | None = None,
         backward: Callable[[Any], Any] | None = None,
         self_strict: bool | None = None,
         other_strict: bool | None = None) -> None:
    """Bind the property of one object to the property of another object.

    The binding works both ways, from the first object to the second and from the second to the first.
    The update happens immediately and whenever a value changes.
    The backward binding takes precedence for the initial synchronization.

    :param self_obj: First object to bind.
    :param self_name: The name of the first property to bind.
    :param other_obj: The second object to bind.
    :param other_name: The name of the second property to bind.
    :param forward: A function to apply to the value before applying it to the second object (default: identity).
    :param backward: A function to apply to the value before applying it to the first object (default: identity).
    :param self_strict: Whether to check (and raise) if the first object has the specified property (default: None,
        performs a check if the object is not a dictionary, *added in version 3.0.0*).
    :param other_strict: Whether to check (and raise) if the second object has the specified property (default: None,
        performs a check if the object is not a dictionary, *added in version 3.0.0*).
    """
    _check_self_and_other_attribute(self_obj, self_name, other_obj, other_name, self_strict, other_strict)
    bind_from(self_obj, self_name, other_obj, other_name, backward=backward, self_strict=False, other_strict=False)
    bind_to(self_obj, self_name, other_obj, other_name, forward=forward, self_strict=False, other_strict=False)


class BindableProperty:

    def __init__(self, on_change: Callable[..., Any] | None = None) -> None:
        self._change_handler = on_change

    def __set_name__(self, _, name: str) -> None:
        self.name = name  # pylint: disable=attribute-defined-outside-init

    def __get__(self, owner: Any, _=None) -> Any:
        return getattr(owner, '___' + self.name)

    def __set__(self, owner: Any, value: Any) -> None:
        has_attr = hasattr(owner, '___' + self.name)
        if not has_attr:
            _make_copyable(type(owner))
        value_changed = has_attr and getattr(owner, '___' + self.name) != value
        if has_attr and not value_changed:
            return
        setattr(owner, '___' + self.name, value)
        key = (id(owner), str(self.name))
        bindable_properties[key] = owner
        _propagate(owner, self.name)
        if value_changed and self._change_handler is not None:
            self._change_handler(owner, value)


def remove(objects: Iterable[Any]) -> None:
    """Remove all bindings that involve the given objects.

    :param objects: The objects to remove.
    """
    object_ids = set(map(id, objects))
    active_links[:] = [
        (source_obj, source_name, target_obj, target_name, transform)
        for source_obj, source_name, target_obj, target_name, transform in active_links
        if id(source_obj) not in object_ids and id(target_obj) not in object_ids
    ]
    for key, binding_list in list(bindings.items()):
        binding_list[:] = [
            (source_obj, target_obj, target_name, transform)
            for source_obj, target_obj, target_name, transform in binding_list
            if id(source_obj) not in object_ids and id(target_obj) not in object_ids
        ]
        if not binding_list:
            del bindings[key]
    for obj_id, name in list(bindable_properties):
        if obj_id in object_ids:
            del bindable_properties[(obj_id, name)]


def reset() -> None:
    """Clear all bindings.

    This function is intended for testing purposes only.
    """
    bindings.clear()
    bindable_properties.clear()
    active_links.clear()


@dataclass_transform()
def bindable_dataclass(cls: TC | None = None, /, *,
                       bindable_fields: Iterable[str] | None = None,
                       **kwargs: Any) -> type[DataclassInstance] | IdentityFunction:
    """A decorator that transforms a class into a dataclass with bindable fields.

    This decorator extends the functionality of ``dataclasses.dataclass`` by making specified fields bindable.
    If ``bindable_fields`` is provided, only the listed fields are made bindable.
    Otherwise, all fields are made bindable by default.

    *Added in version 2.11.0*

    :param cls: class to be transformed into a dataclass
    :param bindable_fields: optional list of field names to make bindable (defaults to all fields)
    :param kwargs: optional keyword arguments to be forwarded to ``dataclasses.dataclass``.
    Usage of ``slots=True`` and ``frozen=True`` are not supported and will raise a ValueError.

    :return: resulting dataclass type
    """
    if cls is None:
        def wrap(cls_):
            return bindable_dataclass(cls_, bindable_fields=bindable_fields, **kwargs)
        return wrap

    for unsupported_option in ('slots', 'frozen'):
        if kwargs.get(unsupported_option):
            raise ValueError(f'`{unsupported_option}=True` is not supported with bindable_dataclass')

    dataclass: type[DataclassInstance] = dataclasses.dataclass(**kwargs)(cls)
    field_names = {field.name for field in dataclasses.fields(dataclass)}
    if bindable_fields is None:
        bindable_fields = field_names
    for field_name in bindable_fields:
        if field_name not in field_names:
            raise ValueError(f'"{field_name}" is not a dataclass field')
        bindable_property = BindableProperty()
        bindable_property.__set_name__(dataclass, field_name)
        setattr(dataclass, field_name, bindable_property)
    return dataclass


def _make_copyable(cls: type[T]) -> None:
    """Tell the copy module to update the ``bindable_properties`` dictionary when an object is copied."""
    if cls in copyreg.dispatch_table:
        return

    def _pickle_function(obj: T) -> tuple[Callable[..., T], tuple[Any, ...]]:
        reduced = obj.__reduce__()
        assert isinstance(reduced, tuple)
        creator = reduced[0]

        def creator_with_hook(*args, **kwargs) -> T:
            copy = creator(*args, **kwargs)
            for attr_name in dir(obj):
                if (id(obj), attr_name) in bindable_properties:
                    bindable_properties[(id(copy), attr_name)] = copy
            return copy
        return (creator_with_hook, *reduced[1:])
    copyreg.pickle(cls, _pickle_function)



from typing import Any

from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Editor(ValueElement, DisableableElement, component='editor.js', default_classes='nicegui-editor'):
    VALUE_PROP: str = 'value'
    LOOPBACK = False

    @resolve_defaults
    def __init__(self,
                 *,
                 placeholder: str | None = DEFAULT_PROP | None,
                 value: str = DEFAULT_PROP | '',
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Editor

        A WYSIWYG editor based on `Quasar's QEditor <https://quasar.dev/vue-components/editor>`_.
        The value is a string containing the formatted text as HTML code.

        :param value: initial value
        :param on_change: callback to be invoked when the value changes
        """
        super().__init__(value=value, on_value_change=on_change)
        self._props.set_optional('placeholder', placeholder)

    def _handle_value_change(self, value: Any) -> None:
        super()._handle_value_change(value)
        if self._send_update_on_value_change:
            self.run_method('updateValue')



import asyncio
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Literal

from typing_extensions import Self

from ... import binding
from ...defaults import DEFAULT_PROP, resolve_defaults
from ...element import Element
from ...events import (
    GenericEventArguments,
    Handler,
    SceneClickEventArguments,
    SceneClickHit,
    SceneDragEventArguments,
    handle_event,
)
from .scene_object3d import Object3D


@dataclass(kw_only=True, slots=True)
class SceneCamera:
    type: Literal['perspective', 'orthographic']
    params: dict[str, float]
    x: float = 0
    y: float = -3
    z: float = 5
    look_at_x: float = 0
    look_at_y: float = 0
    look_at_z: float = 0
    up_x: float = 0
    up_y: float = 0
    up_z: float = 1


@dataclass(kw_only=True, slots=True)
class SceneObject:
    id: str = 'scene'


class Scene(Element, component='scene.js', esm={'nicegui-scene': 'dist'}, default_classes='nicegui-scene'):
    # pylint: disable=import-outside-toplevel
    from .scene_objects import AxesHelper as axes_helper
    from .scene_objects import Box as box
    from .scene_objects import Curve as curve
    from .scene_objects import Cylinder as cylinder
    from .scene_objects import Extrusion as extrusion
    from .scene_objects import Gltf as gltf
    from .scene_objects import Group as group
    from .scene_objects import Line as line
    from .scene_objects import PointCloud as point_cloud
    from .scene_objects import QuadraticBezierTube as quadratic_bezier_tube
    from .scene_objects import Ring as ring
    from .scene_objects import Sphere as sphere
    from .scene_objects import SpotLight as spot_light
    from .scene_objects import Stl as stl
    from .scene_objects import Text as text
    from .scene_objects import Text3d as text3d
    from .scene_objects import Texture as texture

    @resolve_defaults
    def __init__(self,
                 width: int = DEFAULT_PROP | 400,
                 height: int = DEFAULT_PROP | 300,
                 # DEPRECATED: enforce keyword-only arguments in NiceGUI 4.0
                 grid: bool | tuple[int, int] = DEFAULT_PROP | True,
                 camera: SceneCamera | None = None,
                 on_click: Handler[SceneClickEventArguments] | None = None,
                 click_events: list[str] = DEFAULT_PROP | ['click', 'dblclick'],
                 on_drag_start: Handler[SceneDragEventArguments] | None = None,
                 on_drag_end: Handler[SceneDragEventArguments] | None = None,
                 drag_constraints: str = DEFAULT_PROP | '',
                 background_color: str = DEFAULT_PROP | '#eee',
                 control_type: Literal['orbit', 'trackball', 'map'] = DEFAULT_PROP | 'orbit',
                 fps: int = DEFAULT_PROP | 20,
                 show_stats: bool = DEFAULT_PROP | False,
                 ) -> None:
        """3D Scene

        Display a 3D scene using `three.js <https://threejs.org/>`_.
        Currently NiceGUI supports boxes, spheres, cylinders/cones, extrusions, straight lines, curves and textured meshes.
        Objects can be translated, rotated and displayed with different color, opacity or as wireframes.
        They can also be grouped to apply joint movements.

        :param width: width of the canvas
        :param height: height of the canvas
        :param grid: whether to display a grid (boolean or tuple of ``size`` and ``divisions`` for `Three.js' GridHelper <https://threejs.org/docs/#api/en/helpers/GridHelper>`_, default: 100x100)
        :param camera: camera definition, either instance of ``ui.scene.perspective_camera`` (default) or ``ui.scene.orthographic_camera``
        :param on_click: callback to execute when a 3D object is clicked (use ``click_events`` to specify which events to subscribe to)
        :param click_events: list of JavaScript click events to subscribe to (default: ``['click', 'dblclick']``)
        :param on_drag_start: callback to execute when a 3D object is dragged
        :param on_drag_end: callback to execute when a 3D object is dropped
        :param drag_constraints: comma-separated JavaScript expression for constraining positions of dragged objects (e.g. ``'x = 0, z = y / 2'``)
        :param background_color: background color of the scene (default: "#eee")
        :param control_type: type of controls to use for navigating the scene, one of "orbit", "trackball", "map" (default: "orbit", *added in version 3.9.0*)
        :param fps: target frame rate for the scene in frames per second (default: 20, *added in version 3.2.0*)
        :param show_stats: whether to show performance stats (default: ``False``, *added in version 3.2.0*)
        """
        super().__init__()
        self._props['width'] = width
        self._props['height'] = height
        self._props['fps'] = fps
        self._props['show-stats'] = show_stats
        self._props['grid'] = grid
        self._props['background-color'] = background_color
        self.camera = camera or self.perspective_camera()
        self._props['camera-type'] = self.camera.type
        self._props['camera-params'] = self.camera.params
        self.objects: dict[str, Object3D] = {}
        self.stack: list[Object3D | SceneObject] = [SceneObject()]
        self._click_handlers = [on_click] if on_click else []
        self._props['click-events'] = click_events[:]
        self._drag_start_handlers = [on_drag_start] if on_drag_start else []
        self._drag_end_handlers = [on_drag_end] if on_drag_end else []
        self.on('init', self._handle_init)
        self.on('click3d', self._handle_click)
        self.on('dragstart', self._handle_drag)
        self.on('dragend', self._handle_drag)
        self._props['drag-constraints'] = drag_constraints
        self._props['control-type'] = control_type

        self._props.add_rename('background_color', 'background-color')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('camera_params', 'camera-params')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('camera_type', 'camera-type')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('click_events', 'click-events')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('drag_constraints', 'drag-constraints')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('show_stats', 'show-stats')  # DEPRECATED: remove in NiceGUI 4.0

    def on_click(self, callback: Handler[SceneClickEventArguments]) -> Self:
        """Add a callback to be invoked when a 3D object is clicked."""
        self._click_handlers.append(callback)
        return self

    def on_drag_start(self, callback: Handler[SceneDragEventArguments]) -> Self:
        """Add a callback to be invoked when a 3D object is dragged."""
        self._drag_start_handlers.append(callback)
        return self

    def on_drag_end(self, callback: Handler[SceneDragEventArguments]) -> Self:
        """Add a callback to be invoked when a 3D object is dropped."""
        self._drag_end_handlers.append(callback)
        return self

    @staticmethod
    def perspective_camera(*, fov: float = 75, near: float = 0.1, far: float = 1000) -> SceneCamera:
        """Create a perspective camera.

        :param fov: vertical field of view in degrees
        :param near: near clipping plane
        :param far: far clipping plane
        """
        return SceneCamera(type='perspective', params={'fov': fov, 'near': near, 'far': far})

    @staticmethod
    def orthographic_camera(*, size: float = 10, near: float = 0.1, far: float = 1000) -> SceneCamera:
        """Create a orthographic camera.

        The size defines the vertical size of the view volume, i.e. the distance between the top and bottom clipping planes.
        The left and right clipping planes are set such that the aspect ratio matches the viewport.

        :param size: vertical size of the view volume
        :param near: near clipping plane
        :param far: far clipping plane
        """
        return SceneCamera(type='orthographic', params={'size': size, 'near': near, 'far': far})

    def __enter__(self) -> Self:
        Object3D.current_scene = self
        super().__enter__()
        return self

    def __getattribute__(self, name: str) -> Any:
        attribute = super().__getattribute__(name)
        if isinstance(attribute, type) and issubclass(attribute, Object3D):
            Object3D.current_scene = self
        return attribute

    def _handle_init(self) -> None:
        self.move_camera(duration=0)
        self.run_method('init_objects', [obj.data for obj in self.objects.values()])

    async def initialized(self) -> None:
        """Wait until the scene is initialized."""
        event = asyncio.Event()
        self.on('init', event.set, [])
        await self.client.connected()
        await event.wait()

    def _handle_click(self, e: GenericEventArguments) -> None:
        arguments = SceneClickEventArguments(
            sender=self,
            client=self.client,
            click_type=e.args['click_type'],
            button=e.args['button'],
            alt=e.args['alt_key'],
            ctrl=e.args['ctrl_key'],
            meta=e.args['meta_key'],
            shift=e.args['shift_key'],
            hits=[SceneClickHit(
                object_id=hit['object_id'],
                object_name=hit['object_name'],
                x=hit['point']['x'],
                y=hit['point']['y'],
                z=hit['point']['z'],
            ) for hit in e.args['hits']],
        )
        for handler in self._click_handlers:
            handle_event(handler, arguments)

    def _handle_drag(self, e: GenericEventArguments) -> None:
        arguments = SceneDragEventArguments(
            sender=self,
            client=self.client,
            type=e.args['type'],
            object_id=e.args['object_id'],
            object_name=e.args['object_name'],
            x=e.args['x'],
            y=e.args['y'],
            z=e.args['z'],
        )
        if arguments.type == 'dragend':
            self.objects[arguments.object_id].move(arguments.x, arguments.y, arguments.z)

        for handler in (self._drag_start_handlers if arguments.type == 'dragstart' else self._drag_end_handlers):
            handle_event(handler, arguments)

    def __len__(self) -> int:
        return len(self.objects)

    def move_camera(self,
                    x: float | None = None,
                    y: float | None = None,
                    z: float | None = None,
                    look_at_x: float | None = None,
                    look_at_y: float | None = None,
                    look_at_z: float | None = None,
                    up_x: float | None = None,
                    up_y: float | None = None,
                    up_z: float | None = None,
                    duration: float = 0.5) -> None:
        """Move the camera to a new position.

        :param x: camera x position
        :param y: camera y position
        :param z: camera z position
        :param look_at_x: camera look-at x position
        :param look_at_y: camera look-at y position
        :param look_at_z: camera look-at z position
        :param up_x: x component of the camera up vector
        :param up_y: y component of the camera up vector
        :param up_z: z component of the camera up vector
        :param duration: duration of the movement in seconds (default: `0.5`)
        """
        self.camera.x = self.camera.x if x is None else x
        self.camera.y = self.camera.y if y is None else y
        self.camera.z = self.camera.z if z is None else z
        self.camera.look_at_x = self.camera.look_at_x if look_at_x is None else look_at_x
        self.camera.look_at_y = self.camera.look_at_y if look_at_y is None else look_at_y
        self.camera.look_at_z = self.camera.look_at_z if look_at_z is None else look_at_z
        self.camera.up_x = self.camera.up_x if up_x is None else up_x
        self.camera.up_y = self.camera.up_y if up_y is None else up_y
        self.camera.up_z = self.camera.up_z if up_z is None else up_z
        self.run_method('move_camera',
                        self.camera.x, self.camera.y, self.camera.z,
                        self.camera.look_at_x, self.camera.look_at_y, self.camera.look_at_z,
                        self.camera.up_x, self.camera.up_y, self.camera.up_z, duration)

    async def get_camera(self) -> dict[str, Any]:
        """Get the current camera parameters.

        In contrast to the `camera` property,
        the result of this method includes the current camera pose caused by the user navigating the scene in the browser.
        """
        return await self.run_method('get_camera')

    def _handle_delete(self) -> None:
        binding.remove(list(self.objects.values()))
        super()._handle_delete()

    def delete_objects(self, predicate: Callable[[Object3D], bool] = lambda _: True) -> None:
        """Remove objects from the scene.

        :param predicate: function which returns `True` for objects which should be deleted
        """
        for obj in list(self.objects.values()):
            if predicate(obj) and obj.id in self.objects:  # NOTE: object might have been deleted already by its parent
                obj.delete()

    def clear(self) -> Self:
        """Remove all objects from the scene."""
        super().clear()
        self.delete_objects()
        return self



import hashlib
import mimetypes
from collections.abc import AsyncGenerator
from datetime import datetime, timezone
from pathlib import Path

import aiofiles
from fastapi import Request
from fastapi.responses import Response, StreamingResponse

mimetypes.init()

MIN_CHUNK_SIZE = 1024
MAX_CHUNK_SIZE = 8192


def get_range_response(file: Path, request: Request, chunk_size: int) -> Response:
    """Get a Response for the given file, supporting range-requests, E-Tag and Last-Modified."""
    chunk_size = max(MIN_CHUNK_SIZE, min(chunk_size, MAX_CHUNK_SIZE))
    file_size = file.stat().st_size
    last_modified_time = datetime.fromtimestamp(file.stat().st_mtime, timezone.utc)
    start = 0
    end = file_size - 1
    status_code = 200
    e_tag = hashlib.md5((str(last_modified_time) + str(file_size)).encode()).hexdigest()
    if_match_header = request.headers.get('If-None-Match')
    if if_match_header and if_match_header == e_tag:
        return Response(status_code=304)  # Not Modified
    headers = {
        'E-Tag': e_tag,
        'Last-Modified': last_modified_time.strftime(r'%a, %d %b %Y %H:%M:%S GMT'),
    }
    range_header = request.headers.get('range')
    media_type = mimetypes.guess_type(str(file))[0] or 'application/octet-stream'
    if range_header is not None:
        try:
            byte1, byte2 = range_header.split('=')[1].split('-')
            start = int(byte1)
            if byte2:
                end = int(byte2)
        except (IndexError, ValueError):
            return Response(status_code=416, headers={'Content-Range': f'bytes */{file_size}'})
        if start > end or start >= file_size:
            return Response(status_code=416, headers={'Content-Range': f'bytes */{file_size}'})
        end = min(end, file_size - 1)
        status_code = 206  # Partial Content
    content_length = end - start + 1
    headers.update({
        'Content-Length': str(content_length),
        'Content-Range': f'bytes {start}-{end}/{file_size}',
        'Accept-Ranges': 'bytes',
    })

    async def content_reader(file: Path, start: int, end: int) -> AsyncGenerator[bytes, None]:
        async with aiofiles.open(file, 'rb') as data:
            await data.seek(start)
            remaining_bytes = end - start + 1
            while remaining_bytes > 0:
                chunk = await data.read(min(chunk_size, remaining_bytes))
                if not chunk:
                    break
                yield chunk
                remaining_bytes -= len(chunk)
    return StreamingResponse(
        content_reader(file, start, end),
        media_type=media_type,
        headers=headers,
        status_code=status_code,
    )



from typing import Literal

from typing_extensions import Self

from ..element import Element


class Card(Element, default_classes='nicegui-card'):

    def __init__(self, *,
                 align_items: Literal['start', 'end', 'center', 'baseline', 'stretch'] | None = None,
                 ) -> None:
        """Card

        This element is based on Quasar's `QCard <https://quasar.dev/vue-components/card>`_ component.
        It provides a container with a dropped shadow.

        Note:
        In contrast to this element,
        the original QCard has no padding by default and hides outer borders and shadows of nested elements.
        If you want the original behavior, use the `tight` method.

        *Updated in version 2.0.0: Don't hide outer borders and shadows of nested elements anymore.*

        :param align_items: alignment of the items in the card ("start", "end", "center", "baseline", or "stretch"; default: `None`)
        """
        super().__init__('q-card')
        if align_items:
            self._classes.append(f'items-{align_items}')

    def tight(self) -> Self:
        """Remove padding and gaps between nested elements."""
        return self.classes('nicegui-card-tight')


class CardSection(Element):

    def __init__(self) -> None:
        """Card Section

        This element is based on Quasar's `QCardSection <https://quasar.dev/vue-components/card#qcardsection-api>`_ component.
        """
        super().__init__('q-card-section')


class CardActions(Element):

    def __init__(self) -> None:
        """Card Actions

        This element is based on Quasar's `QCardActions <https://quasar.dev/vue-components/card#qcardactions-api>`_ component.
        """
        super().__init__('q-card-actions')



import asyncio
import time

from .button import Button as button
from .markdown import Markdown as markdown
from .markdown import remove_indentation
from .mixins.content_element import ContentElement
from .timer import Timer as timer


class Code(ContentElement, component='code.js', default_classes='nicegui-code'):

    def __init__(self, content: str = '', *, language: str | None = 'python') -> None:
        """Code

        This element displays a code block with syntax highlighting.

        In secure environments (HTTPS or localhost), a copy button is displayed to copy the code to the clipboard.

        :param content: code to display
        :param language: language of the code (default: "python")
        """
        super().__init__(content=remove_indentation(content))

        with self:
            self.markdown = markdown().classes('overflow-auto h-full') \
                .bind_content_from(self, 'content', lambda content: f'```{language}\n{content}\n```')
            self.copy_button = button(icon='content_copy', on_click=self.show_checkmark) \
                .props('round flat size=sm').classes('absolute right-2 top-2 opacity-20 hover:opacity-80') \
                .on('click', js_handler=f'() => navigator.clipboard.writeText(getElement("{self.id}").content)')

        self._props['content'] = self.content
        self._last_scroll: float = 0.0
        self.markdown.on('scroll', self._handle_scroll)
        with self:
            timer(0.1, self._update_copy_button)

    async def show_checkmark(self) -> None:
        """Show a checkmark icon for 3 seconds."""
        self.copy_button.props('icon=check')
        await asyncio.sleep(3.0)
        self.copy_button.props('icon=content_copy')

    def _handle_scroll(self) -> None:
        self._last_scroll = time.time()

    def _update_copy_button(self) -> None:
        self.copy_button.set_visibility(time.time() > self._last_scroll + 1.0)

    def _handle_content_change(self, content: str) -> None:
        self._props['content'] = content



from collections.abc import Callable

from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .input import Input


class Textarea(Input, component='input.js'):

    @resolve_defaults
    def __init__(self,
                 label: str | None = DEFAULT_PROP | None, *,
                 placeholder: str | None = DEFAULT_PROP | None,
                 value: str = DEFAULT_PROPS['model-value'] | '',
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 validation: Callable[..., str | None] | dict[str, Callable[..., bool]] | None = None,
                 ) -> None:
        """Textarea

        This element is based on Quasar's `QInput <https://quasar.dev/vue-components/input>`_ component.
        The ``type`` is set to ``textarea`` to create a multi-line text input.

        You can use the `validation` parameter to define a dictionary of validation rules,
        e.g. ``{'Too long!': lambda value: len(value) < 3}``.
        The key of the first rule that fails will be displayed as an error message.
        Alternatively, you can pass a callable that returns an optional error message.
        To disable the automatic validation on every value change, you can use the `without_auto_validation` method.

        :param label: displayed name for the textarea
        :param placeholder: text to show if no value is entered
        :param value: the initial value of the field
        :param on_change: callback to execute when the value changes
        :param validation: dictionary of validation rules or a callable that returns an optional error message (default: None for no validation)
        """
        super().__init__(label, placeholder=placeholder, value=value, on_change=on_change, validation=validation)
        self._props['type'] = 'textarea'



from typing import Literal

from typing_extensions import Self

from ..binding import BindableProperty
from ..defaults import DEFAULT_PROP, resolve_defaults
from ..element import Element
from ..events import (
    GenericEventArguments,
    Handler,
    KeyboardAction,
    KeyboardKey,
    KeyboardModifiers,
    KeyEventArguments,
    handle_event,
)


class Keyboard(Element, component='keyboard.js'):
    active = BindableProperty()

    @resolve_defaults
    def __init__(self,
                 on_key: Handler[KeyEventArguments] | None = None, *,
                 active: bool = True,
                 repeating: bool = DEFAULT_PROP | True,
                 ignore: list[Literal['input', 'select', 'button', 'textarea']] = DEFAULT_PROP | [
                     'input', 'select', 'button', 'textarea',
                 ],
                 ) -> None:
        """Keyboard

        Adds global keyboard event tracking.

        The ``on_key`` callback receives a ``KeyEventArguments`` object with the following attributes:

        - ``sender``: the ``Keyboard`` element
        - ``client``: the client object
        - ``action``: a ``KeyboardAction`` object with the following attributes:
            - ``keydown``: whether the key was pressed
            - ``keyup``: whether the key was released
            - ``repeat``: whether the key event was a repeat
        - ``key``: a ``KeyboardKey`` object with the following attributes:
            - ``name``: the name of the key (e.g. "a", "Enter", "ArrowLeft"; see `here <https://developer.mozilla.org/en-US/docs/Web/API/UI_Events/Keyboard_event_key_values>`_ for a list of possible values)
            - ``code``: the code of the key (e.g. "KeyA", "Enter", "ArrowLeft")
            - ``location``: the location of the key (0 for standard keys, 1 for left keys, 2 for right keys, 3 for numpad keys)
        - ``modifiers``: a ``KeyboardModifiers`` object with the following attributes:
            - ``alt``: whether the alt key was pressed
            - ``ctrl``: whether the ctrl key was pressed
            - ``meta``: whether the meta key was pressed
            - ``shift``: whether the shift key was pressed

        For convenience, the ``KeyboardKey`` object also has the following properties:
            - ``is_cursorkey``: whether the key is a cursor (arrow) key
            - ``number``: the integer value of a number key (0-9, ``None`` for other keys)
            - ``backspace``, ``tab``, ``enter``, ``shift``, ``control``, ``alt``, ``pause``, ``caps_lock``, ``escape``, ``space``,
              ``page_up``, ``page_down``, ``end``, ``home``, ``arrow_left``, ``arrow_up``, ``arrow_right``, ``arrow_down``,
              ``print_screen``, ``insert``, ``delete``, ``meta``,
              ``f1``, ``f2``, ``f3``, ``f4``, ``f5``, ``f6``, ``f7``, ``f8``, ``f9``, ``f10``, ``f11``, ``f12``: whether the key is the respective key

        :param on_key: callback to be executed when keyboard events occur.
        :param active: boolean flag indicating whether the callback should be executed or not (default: ``True``)
        :param repeating: boolean flag indicating whether held keys should be sent repeatedly (default: ``True``)
        :param ignore: ignore keys when one of these element types is focussed (default: ``['input', 'select', 'button', 'textarea']``)
        """
        super().__init__()
        self._key_handlers = [on_key] if on_key else []
        self.active = active
        self._props['events'] = ['keydown', 'keyup']
        self._props['repeating'] = repeating
        self._props['ignore'] = ignore[:]
        self.on('key', self._handle_key)

    def _handle_key(self, e: GenericEventArguments) -> None:
        if not self.active:
            return

        action = KeyboardAction(
            keydown=e.args['action'] == 'keydown',
            keyup=e.args['action'] == 'keyup',
            repeat=e.args['repeat'],
        )
        modifiers = KeyboardModifiers(
            alt=e.args['altKey'],
            ctrl=e.args['ctrlKey'],
            meta=e.args['metaKey'],
            shift=e.args['shiftKey'],
        )
        key = KeyboardKey(
            name=e.args['key'],
            code=e.args['code'],
            location=e.args['location'],
        )
        arguments = KeyEventArguments(
            sender=self,
            client=self.client,
            action=action,
            modifiers=modifiers,
            key=key,
        )
        for handler in self._key_handlers:
            handle_event(handler, arguments)

    def on_key(self, handler: Handler[KeyEventArguments]) -> Self:
        """Add a callback to be invoked when keyboard events occur."""
        self._key_handlers.append(handler)
        return self



from .persistent_dict import PersistentDict


class PseudoPersistentDict(PersistentDict):
    """A persistent dict that is not actually persistent, but only keeps the data in memory (for internal use only)."""

    async def initialize(self) -> None:
        pass

    def initialize_sync(self) -> None:
        pass



from __future__ import annotations

import asyncio
import weakref
from collections.abc import Iterator
from typing import TYPE_CHECKING, ClassVar

from typing_extensions import Self

from .logging import log

if TYPE_CHECKING:
    from .element import Element


class Slot:
    stacks: ClassVar[dict[int, list[Slot]]] = {}
    '''Maps asyncio task IDs to slot stacks, which keep track of the current slot in each task.'''

    def __init__(self, parent: Element, name: str, template: str | None = None) -> None:
        self.name = name
        self._parent = weakref.ref(parent)
        self.template = template
        self.children: list[Element] = []

    @property
    def parent(self) -> Element:
        """The parent element this slot belongs to."""
        parent = self._parent()
        if parent is None:
            raise RuntimeError('The parent element this slot belongs to has been deleted.')
        return parent

    def __enter__(self) -> Self:
        self.get_stack().append(self)
        return self

    def __exit__(self, *_) -> None:
        self.get_stack().pop()
        self.prune_stack()

    def __iter__(self) -> Iterator[Element]:
        return iter(self.children)

    @classmethod
    def get_stack(cls) -> list[Slot]:
        """Return the slot stack of the current asyncio task."""
        task_id = get_task_id()
        if task_id not in cls.stacks:
            cls.stacks[task_id] = []
        return cls.stacks[task_id]

    @classmethod
    def prune_stack(cls) -> None:
        """Remove the current slot stack if it is empty."""
        task_id = get_task_id()
        if not cls.stacks[task_id]:
            del cls.stacks[task_id]

    @classmethod
    async def prune_stacks(cls) -> None:
        """Remove stale slot stacks."""
        try:
            running = {id(task) for task in asyncio.tasks.all_tasks() if not task.done() and not task.cancelled()}

            # ID 0 is the special ID used in cases where no task is active (see get_task_id).
            # Since it has no associated task, we always assume it to be active
            # in order to avoid pruning the associated stack while the slot is active.
            running.add(0)

            stale_ids = [task_id for task_id in cls.stacks if task_id not in running]
            for task_id in stale_ids:
                del cls.stacks[task_id]
        except Exception:
            log.exception('Error while pruning slot stacks')


def get_task_id() -> int:
    """Return the ID of the current asyncio task."""
    try:
        return id(asyncio.current_task())
    except RuntimeError:
        return 0



from typing import Any


class UserNotify:

    def __init__(self) -> None:
        self.messages: list[str] = []

    def __call__(self, message: str, **kwargs) -> None:
        self.messages.append(message)

    def contains(self, needle: Any) -> bool:
        """Check if any of the messages contain the given substring."""
        return isinstance(needle, str) and any(needle in message for message in self.messages)



import asyncio
from contextlib import suppress

from .. import background_tasks, core, json, optional_features
from ..logging import log
from .persistent_dict import PersistentDict

with suppress(ImportError):
    import redis as redis_sync
    import redis.asyncio as redis
    optional_features.register('redis')


class RedisPersistentDict(PersistentDict):

    def __init__(self, *,
                 url: str,
                 id: str,  # pylint: disable=redefined-builtin
                 key_prefix: str = 'nicegui:',
                 ttl: int | None = None
                 ) -> None:
        if not optional_features.has('redis'):
            raise ImportError('Redis is not installed. Please run "pip install nicegui[redis]".')
        self.url = url
        self._redis_client_params = {
            'health_check_interval': 10,
            'socket_connect_timeout': 5,
            'retry_on_timeout': True,
            **({'socket_keepalive': True} if not url.startswith('unix://') else {}),
        }
        self.redis_client = redis.from_url(self.url, **self._redis_client_params)
        self.pubsub = self.redis_client.pubsub()
        self.key = key_prefix + id
        self.ttl = ttl
        self._listener_task: asyncio.Task | None = None
        super().__init__(data={}, on_change=self.publish)

    async def initialize(self) -> None:
        """Load initial data from Redis and start listening for changes."""
        try:
            data = await self.redis_client.get(self.key)
            self.update(json.loads(data) if data else {})
            self._start_listening()
        except Exception:
            log.warning(f'Could not load data from Redis with key {self.key}')

    def initialize_sync(self) -> None:
        """Load initial data from Redis and start listening for changes in a synchronous context."""
        with redis_sync.from_url(self.url, **self._redis_client_params) as redis_client_sync:
            try:
                data = redis_client_sync.get(self.key)
                self.update(json.loads(data) if data else {})
                self._start_listening()
            except Exception:
                log.warning(f'Could not load data from Redis with key {self.key}')

    def _start_listening(self) -> None:
        async def listen():
            try:
                await self.pubsub.subscribe(self.key + 'changes')
                async for message in self.pubsub.listen():
                    t = message['type']
                    if t == 'message':
                        new_data = json.loads(message['data'])
                        if new_data != self:
                            self.update(new_data)
                    elif t in ('unsubscribe', 'punsubscribe') and message.get('data') == 0:
                        break
            except asyncio.CancelledError:
                pass  # Expected during close()
            except Exception:
                log.exception(f'Unexpected error in Redis listener for {self.key}')
            finally:
                if self.pubsub.subscribed:
                    await self.pubsub.unsubscribe()

        def create_listener_task() -> None:
            self._listener_task = background_tasks.create(listen(), name=f'redis-listen-{self.key}')

        if core.loop and core.loop.is_running():
            create_listener_task()
        else:
            core.app.on_startup(create_listener_task)

    def publish(self) -> None:
        """Publish the data to Redis and notify other instances."""
        async def backup() -> None:
            if not await self.redis_client.exists(self.key) and not self:
                return
            pipeline = self.redis_client.pipeline()
            data = json.dumps(self)
            pipeline.set(self.key, data, ex=self.ttl)
            pipeline.publish(self.key + 'changes', data)
            await pipeline.execute()
        if core.loop:
            background_tasks.create_lazy(backup(), name=f'redis-{self.key}')
        else:
            core.app.on_startup(backup())

    async def close(self) -> None:
        """Close Redis connection and subscription."""
        if self._listener_task and not self._listener_task.done():
            self._listener_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._listener_task
        await self.pubsub.aclose()
        await self.redis_client.aclose()

    def clear(self) -> None:
        super().clear()
        if core.loop:
            background_tasks.create_lazy(self.redis_client.delete(self.key), name=f'redis-delete-{self.key}')
        else:
            core.app.on_startup(self.redis_client.delete(self.key))



from collections.abc import Callable
from typing import Any

from ..context import context


def on_exception(handler: Callable[[Exception], Any] | Callable[[], Any]) -> None:
    """Register a handler for in-page exceptions (after the page has been sent to the browser).

    The callback has an optional parameter of ``Exception``.

    *Added in version 3.6.0*

    :param handler: callback that is called upon occurrence of the event.
    """
    context.client.on_exception(handler)



from __future__ import annotations

from collections import defaultdict
from threading import Thread

from .. import core
from ..events import Handler, NativeEventArguments, handle_event
from . import native


class EventManager:
    """Receives pywebview window events from the subprocess and dispatches them to registered handlers."""

    def __init__(self) -> None:
        self._handlers: dict[str, list[Handler[NativeEventArguments]]] = defaultdict(list)
        self._thread: Thread | None = None

    def on(self, event_type: str, handler: Handler[NativeEventArguments]) -> None:
        """Register a handler for a native window event."""
        self._handlers[event_type].append(handler)

    def start(self) -> None:
        """Start the event listener thread."""
        self._thread = Thread(target=self._event_loop, daemon=True)
        self._thread.start()

    def stop(self) -> None:
        """Stop the event listener thread."""
        if native.event_sender is not None:
            try:
                native.event_sender.send(None)
            except OSError:
                pass  # the pipe might already be closed

    def _dispatch(self, data: dict) -> None:
        args = NativeEventArguments(type=data['type'], args=data.get('args', {}))
        for handler in self._handlers.get(args.type, []):
            handle_event(handler, args)

    def _event_loop(self) -> None:
        assert native.event_receiver is not None
        while True:
            try:
                data = native.event_receiver.recv()
            except (EOFError, OSError, TypeError):
                break  # the pipe was closed
            if data is None:
                break
            if core.loop and core.loop.is_running():
                core.loop.call_soon_threadsafe(self._dispatch, data)
            else:
                self._dispatch(data)


event_manager = EventManager()



# pylint: disable=unused-import
from .general_fixtures import nicegui_reset_globals, pytest_addoption, pytest_configure  # noqa: F401
from .screen_plugin import (  # noqa: F401
    nicegui_chrome_options,
    nicegui_driver,
    nicegui_remove_all_screenshots,
    pytest_runtest_makereport,
    screen,
)
from .user_plugin import create_user, user  # noqa: F401



from __future__ import annotations

from typing import Any, cast

from ..context import context
from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Carousel(ValueElement):

    @resolve_defaults
    def __init__(self, *,
                 value: str | CarouselSlide | None = DEFAULT_PROPS['model-value'] | None,
                 on_value_change: Handler[ValueChangeEventArguments] | None = None,
                 animated: bool = DEFAULT_PROP | False,
                 arrows: bool = DEFAULT_PROP | False,
                 navigation: bool = DEFAULT_PROP | False,
                 ) -> None:
        """Carousel

        This element represents `Quasar's QCarousel <https://quasar.dev/vue-components/carousel#qcarousel-api>`_ component.
        It contains individual carousel slides.

        :param value: `ui.carousel_slide` or name of the slide to be initially selected (default: `None` meaning the first slide)
        :param on_value_change: callback to be executed when the selected slide changes
        :param animated: whether to animate slide transitions (default: `False`)
        :param arrows: whether to show arrows for manual slide navigation (default: `False`)
        :param navigation: whether to show navigation dots for manual slide navigation (default: `False`)
        """
        super().__init__(tag='q-carousel', value=value, on_value_change=on_value_change)
        self._props['animated'] = animated
        self._props['arrows'] = arrows
        self._props['navigation'] = navigation

    def _value_to_model_value(self, value: Any) -> Any:
        return value.props['name'] if isinstance(value, CarouselSlide) else value

    def _handle_value_change(self, value: Any) -> None:
        super()._handle_value_change(value)
        names = [slide.props['name'] for slide in self.default_slot]
        for i, slide in enumerate(self):
            done = i < names.index(value) if value in names else False
            slide.props(f':done={done}')

    def next(self) -> None:
        """Show the next slide."""
        self.run_method('next')

    def previous(self) -> None:
        """Show the previous slide."""
        self.run_method('previous')


class CarouselSlide(DisableableElement, default_classes='nicegui-carousel-slide'):

    def __init__(self, name: str | None = None) -> None:
        """Carousel Slide

        This element represents `Quasar's QCarouselSlide <https://quasar.dev/vue-components/carousel#qcarouselslide-api>`_ component.
        It is a child of a `ui.carousel` element.

        :param name: name of the slide (will be the value of the `ui.carousel` element, auto-generated if `None`)
        """
        super().__init__(tag='q-carousel-slide')
        self.carousel = cast(ValueElement, context.slot.parent)
        name = name or f'slide_{len(self.carousel.default_slot.children)}'
        self._props['name'] = name
        if self.carousel.value is None:
            self.carousel.value = name



from ..defaults import DEFAULT_PROPS, resolve_defaults
from ..events import ClickEventArguments, Handler
from .context_menu import ContextMenu
from .item import Item
from .mixins.value_element import ValueElement


class Menu(ValueElement):

    @resolve_defaults
    def __init__(self, *, value: bool = DEFAULT_PROPS['model-value'] | False) -> None:
        """Menu

        Creates a menu based on Quasar's `QMenu <https://quasar.dev/vue-components/menu>`_ component.
        The menu should be placed inside the element where it should be shown.

        Advanced tip:
        Use the `auto-close` prop to automatically close the menu on any click event directly without a server round-trip.

        :param value: whether the menu is already opened (default: `False`)
        """
        super().__init__(tag='q-menu', value=value)

        # https://github.com/zauberzeug/nicegui/issues/1738
        self._props.add_warning('touch-position',
                                'The prop "touch-position" is not supported by `ui.menu`. '
                                'Use "ui.context_menu()" instead.')

    def open(self) -> None:
        """Open the menu."""
        self.value = True

    def close(self) -> None:
        """Close the menu."""
        self.value = False

    def toggle(self) -> None:
        """Toggle the menu."""
        self.value = not self.value


class MenuItem(Item):

    def __init__(self,
                 text: str = '',
                 on_click: Handler[ClickEventArguments] | None = None, *,
                 auto_close: bool = True,
                 ) -> None:
        """Menu Item

        A menu item to be added to a menu.
        This element is based on Quasar's `QItem <https://quasar.dev/vue-components/list-and-list-items#qitem-api>`_ component.

        :param text: label of the menu item
        :param on_click: callback to be executed when selecting the menu item
        :param auto_close: whether the menu should be closed after a click event (default: `True`)
        """
        super().__init__(text=text, on_click=on_click)

        self._props['clickable'] = True

        self.menu = next((e for e in self.ancestors() if isinstance(e, (Menu, ContextMenu))), None)
        if self.menu and auto_close:
            self.on_click(self.menu.close)



from typing import Literal

from typing_extensions import Self

from ..defaults import DEFAULT_PROPS, resolve_defaults
from ..events import Handler, SlideEventArguments, handle_event
from ..slot import Slot
from .item import Item
from .label import Label
from .mixins.disableable_element import DisableableElement

SlideSide = Literal['left', 'right', 'top', 'bottom']


class SlideItem(DisableableElement):

    def __init__(self, text: str = '', *, on_slide: Handler[SlideEventArguments] | None = None) -> None:
        """Slide Item

        This element is based on Quasar's `QSlideItem <https://quasar.dev/vue-components/slide-item/>`_ component.

        If the ``text`` parameter is provided, a nested ``ui.item`` element will be created with the given text.
        If you want to customize how the text is displayed, you can place custom elements inside the slide item.

        To fill slots for individual slide actions, use the ``left``, ``right``, ``top``, or ``bottom`` methods or
        the ``action`` method with a side argument ("left", "right", "top", or "bottom").

        Once a slide action has occurred, the slide item can be reset back to its initial state using the ``reset`` method.

        *Added in version 2.12.0*

        :param text: text to be displayed (default: "")
        :param on_slide: callback which is invoked when any slide action is activated
        """
        super().__init__(tag='q-slide-item')

        if text:
            with self:
                Item(text)

        if on_slide:
            self.on_slide(None, on_slide)

    def action(self,
               side: SlideSide,
               text: str = '', *,
               on_slide: Handler[SlideEventArguments] | None = None,
               color: str | None = 'primary',
               ) -> Slot:
        """Add a slide action to a specified side.

        :param side: side of the slide item where the slide should be added ("left", "right", "top", "bottom")
        :param text: text to be displayed (default: "")
        :param on_slide: callback which is invoked when the slide action is activated
        :param color: the color of the slide background (either a Quasar, Tailwind, or CSS color or ``None``, default: "primary")
        """
        self._props.set_optional(f'{side}-color', color)

        if on_slide:
            self.on_slide(side, on_slide)

        slot = self.add_slot(side)
        if text:
            with slot:
                Label(text)

        return slot

    @resolve_defaults
    def left(self,
             text: str = '', *,
             on_slide: Handler[SlideEventArguments] | None = None,
             color: str | None = DEFAULT_PROPS['left-color'] | 'primary',
             ) -> Slot:
        """Add a slide action to the left side.

        :param text: text to be displayed (default: "")
        :param on_slide: callback which is invoked when the slide action is activated
        :param color: the color of the slide background (either a Quasar, Tailwind, or CSS color or ``None``, default: "primary")
        """
        return self.action('left', text=text, on_slide=on_slide, color=color)

    @resolve_defaults
    def right(self,
              text: str = '', *,
              on_slide: Handler[SlideEventArguments] | None = None,
              color: str | None = DEFAULT_PROPS['right-color'] | 'primary',
              ) -> Slot:
        """Add a slide action to the right side.

        :param text: text to be displayed (default: "")
        :param on_slide: callback which is invoked when the slide action is activated
        :param color: the color of the slide background (either a Quasar, Tailwind, or CSS color or ``None``, default: "primary")
        """
        return self.action('right', text=text, on_slide=on_slide, color=color)

    @resolve_defaults
    def top(self,
            text: str = '', *,
            on_slide: Handler[SlideEventArguments] | None = None,
            color: str | None = DEFAULT_PROPS['top-color'] | 'primary',
            ) -> Slot:
        """Add a slide action to the top side.

        :param text: text to be displayed (default: "")
        :param on_slide: callback which is invoked when the slide action is activated
        :param color: the color of the slide background (either a Quasar, Tailwind, or CSS color or ``None``, default: "primary")
        """
        return self.action('top', text=text, on_slide=on_slide, color=color)

    @resolve_defaults
    def bottom(self,
               text: str = '', *,
               on_slide: Handler[SlideEventArguments] | None = None,
               color: str | None = DEFAULT_PROPS['bottom-color'] | 'primary',
               ) -> Slot:
        """Add a slide action to the bottom side.

        :param text: text to be displayed (default: "")
        :param on_slide: callback which is invoked when the slide action is activated
        :param color: the color of the slide background (either a Quasar, Tailwind, or CSS color or ``None``, default: "primary")
        """
        return self.action('bottom', text=text, on_slide=on_slide, color=color)

    def on_slide(self, side: SlideSide | None, handler: Handler[SlideEventArguments]) -> Self:
        """Add a callback to be invoked when the slide action is activated."""
        self.on(side or 'action', lambda e: handle_event(handler, SlideEventArguments(sender=self,
                                                                                      client=self.client,
                                                                                      side=e.args.get('side', side))))
        return self

    def reset(self) -> None:
        """Reset the slide item to its initial state."""
        self.run_method('reset')



import gc
from collections.abc import Callable
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, Literal

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.gzip import GZipMiddleware
from starlette.types import ASGIApp

from . import core, helpers, storage
from .air import Air
from .language import Language
from .logging import log
from .middlewares import RedirectWithPrefixMiddleware, SetCacheControlMiddleware
from .nicegui import _shutdown, _startup
from .server import Server


def run_with(
    app: FastAPI, *,
    root: Callable | None = None,
    title: str = 'NiceGUI',
    viewport: str = 'width=device-width, initial-scale=1',
    favicon: str | Path | None = None,
    dark: bool | None = False,
    language: Language = 'en-US',
    binding_refresh_interval: float | None = 0.1,
    reconnect_timeout: float = 3.0,
    message_history_length: int = 1000,
    cache_control_directives: str = 'public, max-age=31536000, immutable, stale-while-revalidate=31536000',
    gzip_middleware_factory: Callable[[ASGIApp], GZipMiddleware] | None = GZipMiddleware,
    mount_path: str = '/',
    on_air: str | Literal[True] | None = None,
    tailwind: bool = True,
    unocss: Literal['mini', 'wind3', 'wind4'] | None = None,
    prod_js: bool = True,
    storage_secret: str | None = None,
    session_middleware_kwargs: dict[str, Any] | None = None,
    show_welcome_message: bool = True,
) -> None:
    """Run NiceGUI with FastAPI.

    :param app: FastAPI app
    :param root: root page function (*added in version 3.0.0*)
    :param title: page title (default: `'NiceGUI'`, can be overwritten per page)
    :param viewport: page meta viewport content (default: `'width=device-width, initial-scale=1'`, can be overwritten per page)
    :param favicon: relative filepath, absolute URL to a favicon (default: `None`, NiceGUI icon will be used) or emoji (e.g. `'🚀'`, works for most browsers)
    :param dark: whether to use Quasar's dark mode (default: `False`, use `None` for "auto" mode)
    :param language: language for Quasar elements (default: `'en-US'`)
    :param binding_refresh_interval: interval for updating active links (default: 0.1 seconds, bigger is more CPU friendly, *since version 3.4.0*: can be ``None`` to disable update loop)
    :param reconnect_timeout: maximum time the server waits for the browser to reconnect (default: 3.0 seconds)
    :param message_history_length: maximum number of messages that will be stored and resent after a connection interruption (default: 1000, use 0 to disable, *added in version 2.9.0*)
    :param cache_control_directives: cache control directives for internal static files (default: `'public, max-age=31536000, immutable, stale-while-revalidate=31536000'`)
    :param gzip_middleware_factory: GZipMiddleware factory function (e.g. ``lambda app: GZipMiddleware(app, minimum_size=500, compresslevel=9)``, defaults to Starlette's ``GZipMiddleware``, can be ``None`` to disable, *added in version 3.5.0*)
    :param mount_path: mount NiceGUI at this path (default: `'/'`)
    :param on_air: tech preview: `allows temporary remote access <https://nicegui.io/documentation/section_configuration_deployment#nicegui_on_air>`_ if set to `True` (default: disabled)
    :param tailwind: whether to use Tailwind CSS (experimental, default: `True`)
    :param unocss: use UnoCSS with the specified preset instead of Tailwind CSS (default: ``None``, options: "mini", "wind3", "wind4", *added in version 3.7.0*)
    :param prod_js: whether to use the production version of Vue and Quasar dependencies (default: `True`)
    :param storage_secret: secret key for browser-based storage (default: `None`, a value is required to enable ui.storage.user and ui.storage.browser)
    :param session_middleware_kwargs: additional keyword arguments passed to SessionMiddleware that creates the session cookies used for browser-based storage
    :param show_welcome_message: whether to show the welcome message (default: `True`)
    """
    if core.script_mode:
        log.warning(
            'NiceGUI elements were created outside of a page context before ui.run_with() was called.\n'
            'This is not supported and the elements will be discarded.\n'
            'Move UI element creation into a @ui.page function or an app.on_startup handler.'
        )
        if core.script_client is not None:
            core.script_client.delete()
            core.script_client = None
        core.script_mode = False

    core.app.config.add_run_config(
        reload=False,
        title=title,
        viewport=viewport,
        favicon=favicon,
        dark=dark,
        language=language,
        binding_refresh_interval=binding_refresh_interval,
        reconnect_timeout=reconnect_timeout,
        message_history_length=message_history_length,
        tailwind=tailwind,
        unocss=unocss,
        prod_js=prod_js,
        show_welcome_message=show_welcome_message,
        cache_control_directives=cache_control_directives,
    )
    core.root = root
    storage.set_storage_secret(storage_secret, session_middleware_kwargs, parent_app=app)
    if not helpers.is_pytest() and gzip_middleware_factory is not None:
        core.app.add_middleware(gzip_middleware_factory)
    core.app.add_middleware(RedirectWithPrefixMiddleware)
    core.app.add_middleware(SetCacheControlMiddleware)

    if on_air:
        core.air = Air('' if on_air is True else on_air)

    app.mount(mount_path, core.app)
    main_app_lifespan = app.router.lifespan_context

    @asynccontextmanager
    async def lifespan_wrapper(app):
        def _get_server_instance() -> uvicorn.Server | None:
            for server in (obj for obj in gc.get_objects() if isinstance(obj, uvicorn.Server)):
                wrapped = server.config.loaded_app
                while wrapped is not None:
                    if wrapped is app:
                        return server
                    wrapped = getattr(wrapped, 'app', None)
            return None
        if (instance := _get_server_instance()) is not None:
            Server.instance = instance
        await _startup()
        async with main_app_lifespan(app) as state:
            yield state
        await _shutdown()

    app.router.lifespan_context = lifespan_wrapper



# pylint: disable=C0116
from __future__ import annotations

import inspect
import warnings
from collections.abc import Callable
from multiprocessing import Pipe, Queue
from multiprocessing.connection import Connection
from typing import Any

from .. import run
from ..logging import log

method_queue: Queue | None = None
response_queue: Queue | None = None
event_receiver: Connection | None = None
event_sender: Connection | None = None


def create_queues() -> None:
    """Create the message queues and event pipe. (For internal use only.)"""
    global method_queue, response_queue, event_receiver, event_sender  # pylint: disable=global-statement # noqa: PLW0603
    method_queue = Queue()
    response_queue = Queue()
    event_receiver, event_sender = Pipe(duplex=False)


def remove_queues() -> None:
    """Remove the message queues and event pipe. (For internal use only.)"""
    global method_queue, response_queue, event_receiver, event_sender  # pylint: disable=global-statement # noqa: PLW0603
    if method_queue is not None:
        method_queue.close()
        method_queue.join_thread()
        method_queue = None
    if response_queue is not None:
        response_queue.close()
        response_queue.join_thread()
        response_queue = None
    if event_receiver is not None:
        event_receiver.close()
        event_receiver = None
    if event_sender is not None:
        event_sender.close()
        event_sender = None


try:
    with warnings.catch_warnings():
        # webview depends on bottle which uses the deprecated CGI function (https://github.com/bottlepy/bottle/issues/1403)
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        import webview
        from webview.window import FixPoint

    class WindowProxy(webview.Window):

        def __init__(self) -> None:  # pylint: disable=super-init-not-called
            pass  # NOTE we don't call super().__init__ here because this is just a proxy to the actual window

        async def get_always_on_top(self) -> bool:
            """Get whether the window is always on top."""
            return await self._request()

        def set_always_on_top(self, on_top: bool) -> None:
            """Set whether the window is always on top."""
            self._send(on_top)

        async def get_size(self) -> tuple[int, int]:
            """Get the window size as tuple (width, height)."""
            return await self._request()

        async def get_position(self) -> tuple[int, int]:
            """Get the window position as tuple (x, y)."""
            return await self._request()

        def load_url(self, url: str) -> None:
            self._send(url)

        def load_html(self, html: str, base_uri: str = ...) -> None:  # type: ignore
            self._send(html, base_uri)

        def load_css(self, stylesheet: str) -> None:
            self._send(stylesheet)

        def set_title(self, title: str) -> None:
            self._send(title)

        async def get_cookies(self) -> Any:  # pylint: disable=invalid-overridden-method
            return await self._request()

        async def get_current_url(self) -> str:  # type: ignore # pylint: disable=invalid-overridden-method
            return await self._request()

        def destroy(self) -> None:
            self._send()

        def show(self) -> None:
            self._send()

        def hide(self) -> None:
            self._send()

        def set_window_size(self, width: int, height: int) -> None:
            self._send(width, height)

        def resize(self, width: int, height: int, fix_point: FixPoint = FixPoint.NORTH | FixPoint.WEST) -> None:
            self._send(width, height, fix_point)

        def maximize(self) -> None:
            self._send()

        def minimize(self) -> None:
            self._send()

        def restore(self) -> None:
            self._send()

        def toggle_fullscreen(self) -> None:
            self._send()

        def move(self, x: int, y: int) -> None:
            self._send(x, y)

        async def evaluate_js(  # type: ignore # pylint: disable=arguments-differ,invalid-overridden-method
            self,
            script: str,
        ) -> str:
            return await self._request(script)

        async def create_confirmation_dialog(  # type: ignore # pylint: disable=invalid-overridden-method
            self,
            title: str,
            message: str,
        ) -> bool:
            return await self._request(title, message)

        async def create_file_dialog(  # type: ignore # pylint: disable=invalid-overridden-method
            self,
            dialog_type: int = webview.FileDialog.OPEN if hasattr(webview, 'FileDialog') else webview.OPEN_DIALOG,
            directory: str = '',
            allow_multiple: bool = False,
            save_filename: str = '',
            file_types: tuple[str, ...] = (),
        ) -> tuple[str, ...] | None:
            return await self._request(
                dialog_type=dialog_type,
                directory=directory,
                allow_multiple=allow_multiple,
                save_filename=save_filename,
                file_types=file_types,
            )

        def expose(self, function: Callable) -> None:  # type: ignore # pylint: disable=arguments-differ
            raise NotImplementedError(f'exposing "{function}" is not supported')

        def _send(self, *args: Any, **kwargs: Any) -> None:
            assert method_queue is not None
            name = inspect.currentframe().f_back.f_code.co_name  # type: ignore
            method_queue.put((name, args, kwargs))

        async def _request(self, *args: Any, **kwargs: Any) -> Any:
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                assert method_queue is not None
                assert response_queue is not None
                try:
                    method_queue.put((name, args, kwargs))
                    return response_queue.get()  # wait for the method to be called and writing its result to the queue
                except Exception:
                    log.exception(f'error in {name}')
                    return None
            name = inspect.currentframe().f_back.f_code.co_name  # type: ignore
            return await run.io_bound(wrapper, *args, **kwargs)

        def signal_server_shutdown(self) -> None:
            """Signal the server shutdown."""
            self._send()

except ModuleNotFoundError:
    class WindowProxy:  # type: ignore
        pass  # just a dummy if webview is not installed



from typing_extensions import Self

from ...defaults import DEFAULT_PROP, resolve_defaults
from ...events import Handler, MermaidNodeClickEventArguments, handle_event
from ..mixins.content_element import ContentElement


class Mermaid(ContentElement, component='mermaid.js', esm={'nicegui-mermaid': 'dist'}):
    CONTENT_PROP = 'content'

    @resolve_defaults
    def __init__(self,
                 content: str,
                 config: dict | None = DEFAULT_PROP | None,
                 *,
                 on_node_click: Handler[MermaidNodeClickEventArguments] | None = None,
                 ) -> None:
        """Mermaid Diagrams

        Renders diagrams and charts written in the Markdown-inspired `Mermaid <https://mermaid.js.org/>`_ language.
        The mermaid syntax can also be used inside Markdown elements by providing the extension string 'mermaid' to the ``ui.markdown`` element.

        The optional configuration dictionary is passed directly to mermaid before the first diagram is rendered.
        This can be used to set such options as

            ``{'securityLevel': 'loose', ...}`` - allow running JavaScript when a node is clicked
            ``{'logLevel': 'info', ...}`` - log debug info to the console

        Refer to the Mermaid documentation for the ``mermaid.initialize()`` method for a full list of options.

        **Note:** When using click handlers with ``securityLevel: 'loose'``,
        make sure to use unique node IDs across diagrams to avoid click events being bound to the wrong elements.

        :param content: the Mermaid content to be displayed
        :param config: configuration dictionary to be passed to ``mermaid.initialize()``
        :param on_node_click: callback that is invoked when a node is clicked (*added in version 3.3.0*)
        """
        super().__init__(content=content)
        self._props['config'] = config

        if on_node_click is not None:
            self.on_node_click(on_node_click)

    def on_node_click(self, callback: Handler[MermaidNodeClickEventArguments]) -> Self:
        """Add a callback to be invoked when a node is clicked."""
        self.on('node_click', lambda e: handle_event(callback, MermaidNodeClickEventArguments(
            sender=self,
            client=self.client,
            node_id='-'.join(e.args.split('-')[1:-1])  # extract Node ID from HTML ID (<type>-<node_id>-<index>)
        )))
        self._props['clickable'] = True
        return self

    def _handle_content_change(self, content: str) -> None:
        self._props[self.CONTENT_PROP] = content.strip()
        self.run_method('update', content.strip())



import os
from functools import lru_cache

from docutils.core import publish_parts

from .markdown import Markdown, remove_indentation


class ReStructuredText(Markdown):

    def __init__(self, content: str = '') -> None:
        """ReStructuredText

        Renders ReStructuredText onto the page.

        :param content: the ReStructuredText content to be displayed
        """
        super().__init__(content=content)

    def _handle_content_change(self, content: str) -> None:
        html = prepare_content(content)
        if self._props.get('innerHTML') != html:
            self._props['innerHTML'] = html


@lru_cache(maxsize=int(os.environ.get('RST_CONTENT_CACHE_SIZE', '1000')))
def prepare_content(content: str) -> str:
    """Render ReStructuredText content to HTML."""
    html = publish_parts(
        remove_indentation(content),
        writer_name='html4',
        settings_overrides={'syntax_highlight': 'short'},
    )
    return html['html_body'].replace('<div class="document"', '<div class="codehilite"')



from collections.abc import Awaitable, Callable
from typing import Any, TypeAlias

from typing_extensions import Self

from ... import background_tasks, helpers
from .value_element import ValueElement

ValidationFunction: TypeAlias = Callable[[Any], str | None | Awaitable[str | None]]
ValidationDict = dict[str, Callable[[Any], bool]]


class ValidationElement(ValueElement):

    def __init__(self, validation: ValidationFunction | ValidationDict | None, **kwargs: Any) -> None:
        self._validation = validation
        self._auto_validation = True
        self._error: str | None = None
        super().__init__(**kwargs)
        self._props['error'] = None if validation is None else False  # NOTE: reserve bottom space for error message

    @property
    def validation(self) -> ValidationFunction | ValidationDict | None:
        """The validation function or dictionary of validation functions."""
        return self._validation

    @validation.setter
    def validation(self, validation: ValidationFunction | ValidationDict | None) -> None:
        """Sets the validation function or dictionary of validation functions.

        :param validation: validation function or dictionary of validation functions (``None`` to disable validation)
        """
        self._validation = validation
        self.validate(return_result=False)

    @property
    def error(self) -> str | None:
        """The latest error message from the validation functions."""
        return self._error

    @error.setter
    def error(self, error: str | None) -> None:
        """Sets the error message.

        :param error: The optional error message
        """
        new_error_prop = None if self.validation is None else (error is not None)
        if self._error == error and self._props['error'] == new_error_prop:
            return
        self._error = error
        self._props['error'] = new_error_prop
        self._props['error-message'] = error

    def validate(self, *, return_result: bool = True) -> bool:
        """Validate the current value and set the error message if necessary.

        For async validation functions, ``return_result`` must be set to ``False`` and the return value will be ``True``,
        independently of the validation result which is evaluated in the background.

        *Updated in version 2.7.0: Added support for async validation functions.*

        :param return_result: whether to return the result of the validation (default: ``True``)
        :return: whether the validation was successful (always ``True`` for async validation functions)
        """
        if helpers.is_coroutine_function(self._validation):
            async def await_error():
                assert callable(self._validation)
                result = self._validation(self.value)
                assert isinstance(result, Awaitable)
                self.error = await result
            if return_result:
                raise NotImplementedError('The validate method cannot return results for async validation functions.')
            background_tasks.create(await_error(), name=f'validate {self.id}')
            return True

        if callable(self._validation):
            result = self._validation(self.value)
            assert not isinstance(result, Awaitable)
            self.error = result
            return self.error is None

        if isinstance(self._validation, dict):
            for message, check in self._validation.items():
                if not check(self.value):
                    self.error = message
                    return False

        self.error = None
        return True

    def without_auto_validation(self) -> Self:
        """Disable automatic validation on value change."""
        self._auto_validation = False
        return self

    def _handle_value_change(self, value: Any) -> None:
        super()._handle_value_change(value)
        if self._auto_validation:
            self.validate(return_result=False)



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Range(ValueElement, DisableableElement):

    @resolve_defaults
    def __init__(self, *,
                 min: float,  # pylint: disable=redefined-builtin
                 max: float,  # pylint: disable=redefined-builtin
                 step: float = DEFAULT_PROP | 1.0,
                 value: dict[str, float] | None = DEFAULT_PROPS['model-value'] | None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Range

        This element is based on Quasar's `QRange <https://quasar.dev/vue-components/range>`_ component.

        :param min: lower bound of the range
        :param max: upper bound of the range
        :param step: step size
        :param value: initial value to set min and max position of the range (default: ``min`` to ``max``)
        :param on_change: callback which is invoked when the user releases the range
        """
        super().__init__(tag='q-range', value=value or {'min': min, 'max': max},
                         on_value_change=on_change, throttle=0.05)
        self._props['min'] = min
        self._props['max'] = max
        self._props['step'] = step

    @property
    def min(self) -> float:
        """The minimum value allowed."""
        return self._props['min']

    @min.setter
    def min(self, value: float) -> None:
        if self._props['min'] == value:
            return
        self._props['min'] = value

    @property
    def max(self) -> float:
        """The maximum value allowed."""
        return self._props['max']

    @max.setter
    def max(self, value: float) -> None:
        if self._props['max'] == value:
            return
        self._props['max'] = value

    @property
    def step(self) -> float:
        """The step size between valid values."""
        return self._props['step']

    @step.setter
    def step(self, value: float) -> None:
        if self._props['step'] == value:
            return
        self._props['step'] = value



from collections.abc import Callable
from contextlib import suppress
from typing import Literal

from typing_extensions import Self

from ... import optional_features
from ...awaitable_response import AwaitableResponse
from ...defaults import DEFAULT_PROP, resolve_defaults
from ...element import Element
from ...events import (
    EChartComponentClickEventArguments,
    EChartPointClickEventArguments,
    GenericEventArguments,
    Handler,
    handle_event,
)

with suppress(ImportError):
    from pyecharts.charts.base import default, json
    from pyecharts.charts.chart import Base as Chart
    from pyecharts.commons.utils import JsCode
    JS_CODE_MARKER = JsCode('\n').js_code.split('\n')[0]
    optional_features.register('pyecharts')


class EChart(Element, component='echart.js', esm={'nicegui-echart': 'dist'}, default_classes='nicegui-echart'):

    @resolve_defaults
    def __init__(self,
                 options: dict,
                 on_point_click: Handler[EChartPointClickEventArguments] | None = None, *,
                 on_click: Handler[EChartComponentClickEventArguments] | None = None,
                 enable_3d: bool = DEFAULT_PROP | False,
                 renderer: Literal['canvas', 'svg'] = DEFAULT_PROP | 'canvas',
                 theme: str | dict | None = DEFAULT_PROP | None,
                 ) -> None:
        """Apache EChart

        An element to create a chart using `ECharts <https://echarts.apache.org/>`_.
        Updates can be pushed to the chart by changing the `options` property.

        :param options: dictionary of EChart options
        :param on_point_click: callback that is invoked when a point is clicked
        :param on_click: callback that is invoked when any component is clicked (*added in version 3.5.0*)
        :param enable_3d: enforce importing the echarts-gl library
        :param renderer: renderer to use ("canvas" or "svg", *added in version 2.7.0*)
        :param theme: an EChart theme configuration (dictionary or a URL returning a JSON object, *added in version 2.15.0*)
        """
        super().__init__()
        self._props['options'] = options
        self._props['enable-3d'] = enable_3d or any('3D' in key for key in options)
        self._props['renderer'] = renderer
        self._props['theme'] = theme
        self._update_method = 'update_chart'

        if on_point_click:
            self.on_point_click(on_point_click)
        if on_click:
            self.on_click(on_click)

        self._props.add_rename('enable_3d', 'enable-3d')  # DEPRECATED: remove in NiceGUI 4.0

    def on_point_click(self, callback: Handler[EChartPointClickEventArguments]) -> Self:
        """Add a callback to be invoked when a point is clicked."""
        def handle_point_click(e: GenericEventArguments) -> None:
            if e.args['componentType'] != 'series':
                return
            handle_event(callback, EChartPointClickEventArguments(
                sender=self,
                client=self.client,
                component_type=e.args['componentType'],
                series_type=e.args['seriesType'],
                series_index=e.args['seriesIndex'],
                series_name=e.args['seriesName'],
                name=e.args['name'],
                data_index=e.args['dataIndex'],
                data=e.args['data'],
                data_type=e.args.get('dataType'),
                value=e.args['value'],
            ))
        self.on('componentClick', handle_point_click, [
            'componentType',
            'seriesType',
            'seriesIndex',
            'seriesName',
            'name',
            'dataIndex',
            'data',
            'dataType',
            'value',
        ])
        return self

    def on_click(self, callback: Handler[EChartComponentClickEventArguments]) -> Self:
        """Add a callback to be invoked when any component is clicked."""
        def handle_click(e: GenericEventArguments) -> None:
            handle_event(callback, EChartComponentClickEventArguments(
                sender=self,
                client=self.client,
                component_type=e.args['componentType'],
                name=e.args.get('name'),
            ))
        self.on('componentClick', handle_click, [
            'componentType',
            'name',
        ])
        return self

    @classmethod
    def from_pyecharts(cls, chart: 'Chart', on_point_click: Callable | None = None) -> Self:
        """Create an echart element from a pyecharts object.

        :param chart: pyecharts chart object
        :param on_click_point: callback which is invoked when a point is clicked

        :return: echart element
        """
        options = json.loads(json.dumps(chart.get_options(), default=default, ignore_nan=True))
        stack = [options]
        while stack:
            current = stack.pop()
            if isinstance(current, list):
                stack.extend(current)
            elif isinstance(current, dict):
                for key, value in tuple(current.items()):
                    if isinstance(value, str) and value.startswith(JS_CODE_MARKER) and value.endswith(JS_CODE_MARKER):
                        current[f':{key}'] = current.pop(key)[len(JS_CODE_MARKER):-len(JS_CODE_MARKER)]
                    else:
                        stack.append(value)
        return cls(options, on_point_click)

    @property
    def options(self) -> dict:
        """The options dictionary."""
        return self._props['options']

    def run_chart_method(self, name: str, *args, timeout: float = 1) -> AwaitableResponse:
        """Run a method of the EChart instance.

        See the `ECharts documentation <https://echarts.apache.org/en/api.html#echartsInstance>`_ for a list of methods.

        If the function is awaited, the result of the method call is returned.
        Otherwise, the method is executed without waiting for a response.

        :param name: name of the method (a prefix ":" indicates that the arguments are JavaScript expressions)
        :param args: arguments to pass to the method (Python objects or JavaScript expressions)
        :param timeout: timeout in seconds (default: 1 second)

        :return: AwaitableResponse that can be awaited to get the result of the method call
        """
        return self.run_method('run_chart_method', name, *args, timeout=timeout)



from __future__ import annotations

import multiprocessing
import multiprocessing.synchronize
import socket
import threading
from typing import Any

import uvicorn

from . import core, storage
from .native import native


class CustomServerConfig(uvicorn.Config):
    storage_secret: str | None = None
    method_queue: multiprocessing.Queue | None = None
    response_queue: multiprocessing.Queue | None = None
    shutdown_event: multiprocessing.synchronize.Event | None = None
    session_middleware_kwargs: dict[str, Any] | None = None


class Server(uvicorn.Server):
    instance: uvicorn.Server

    @classmethod
    def create_singleton(cls, config: CustomServerConfig) -> None:
        """Create a singleton instance of the server."""
        cls.instance = cls(config=config)

    def run(self, sockets: list[socket.socket] | None = None) -> None:
        self.instance = self
        assert isinstance(self.config, CustomServerConfig)
        if self.config.method_queue is not None and self.config.response_queue is not None:
            core.app.native.main_window = native.WindowProxy()
            native.method_queue = self.config.method_queue
            native.response_queue = self.config.response_queue
            if (event := self.config.shutdown_event) is not None:
                def monitor_shutdown_event() -> None:
                    event.wait()
                    self.should_exit = True
                threading.Thread(target=monitor_shutdown_event, daemon=True).start()

        storage.set_storage_secret(self.config.storage_secret, self.config.session_middleware_kwargs)
        super().run(sockets=sockets)
